// $Id: dc_hdt.cc,v 1.14 2000/07/01 04:19:16 rajiyer Exp rajiyer $

// ------------------------------------------------------------- //
// dc_hdt.c:  implementation of internal functions and user      //
//            interface functions for the implementation of      //
//            the dynamic connectivity algorithm by              //
//	      Holm, de Lichtenberg, and Thorup.			 //
//                                                               //
//	      The code is a modified version of David Albert's   //
//	      implementation of the Henzinger-King algorithm.    //
//                                                               //
// Version 1.0 (July, 2000)                                      //
//								 //
// Implementation by Raj D. Iyer and H. S. Rahul		 //
// There is ABSOLUTELY NO WARRANTY.                              //
// ------------------------------------------------------------- //

#define INFINITY 9999999

// Needs to be global :-( because replace can go across multiple
// samp_for_replacement calls and we need to ensure that this is not
// wiped out.
int num_sampled;

#include <dc_hdt.h>
#include <LEDA/queue.h>
#include <sys/time.h>
#include <stdio.h>

int SAMP_THRESHOLD_EDGES, BASE_SIZE;

#ifdef DEBUG
void 
print_list(list<edge> &L, dc_graph *G)
{
    list_item f;

    f = L.first();
    while( f != nil )
    {
        G->print_edge(L.contents(f));
	cout << ", "<<flush;
	f = L.succ( f );
    }
    cout << ".\n"<<flush;
}
#endif

edge
hdt_base::ins(node u, node v)
// create an edge connecting u and v and return it
{
#ifdef STATISTICS
    n_ins++;
#endif

    edge e = Gp->new_edge(u,v);
    (*Gp)[e] = new hdt_edge_struct();

    // test whether u and v are already connected
    if(!connected(u,v,0))
    {   // they are not, so e becomes a forest edge at level 0
        insert_tree(e,0,true);
    }
    else
    {    // insert this non-tree edge at level 0
        insert_non_tree(e,0);
    }
    return e;
}

void 
hdt_base::del(edge e)
// remove the edge e from the data structure, but not from the graph
{
#ifdef STATISTICS
    n_del++;
#endif

    int e_level = level(e);

    // if e is not an edge in F
    if(!tree_edge(e)) 
    {
#ifdef STATISTICS
	n_nontree_deletes_at_level[e_level]++;
#endif

	delete_non_tree(e);
    }
    else  // e is a tree edge
    {
#ifdef STATISTICS
	n_tree_deletes_at_level[e_level]++;
#endif

	// remember e
        node u = source(e);
	node v = target(e);

	// remove e
        delete_tree(e);

	// delete specific information for tree edges stored at e
        for(int j=max_level; j>=0; j--) 
	    delete[] (*Gp)[e]->tree_occ[j];
	delete[] (*Gp)[e]->tree_occ;

	// look for a replacement edge
        replace(u,v,e_level);
    }

    // delete additional edge info
    delete (*Gp)[e];
    Gp->del_edge(e);
}

bool 
hdt_base::connected(node u, node v)
// return true if u and v are connected in the current graph
// and false otherwise
{
#ifdef STATISTICS
    n_query++;
#endif

    return connected(u,v,0);
}

void 
hdt_base::print_statistics(ostream& out)
{
#ifdef STATISTICS
    out << "\nStatistics\n==========\n";
    out << "number of nodes: " << Gp->number_of_nodes();
    out << " final number of edges: " << Gp->number_of_edges();
    out << " final number of components: " << components << "\n\n";
    out << "user supplied operations\n";
    out << "number of ins operations: " << n_ins << "\n";
    out << "number of del operations: " << n_del << "\n";
    out << "number of connected operations: " << n_query << "\n\n";
    out << "internal variables\n";
    out << "number of levels: " << max_level+1 << "\n";
    out << "internal functions\n";
    out << "number of connected operations: " << n_connected << "\n";
    out << "number of insert_tree operations: " << n_ins_tree << "\n";
    out << "number of delete_tree operations: " << n_del_tree << "\n";
    out << "number of insert_non_tree operations: " << n_ins_non_tree << "\n";
    out << "number of delete_non_tree operations: " << n_del_non_tree << "\n";
    out << "number of replace operations: " << n_replace << "\n";
    out << "number of successful replaces: " << n_replace_found << "\n";
    out << "number of tree edges moved up: " << n_tree_moved_up << "\n";
    out << "number of nontree edges moved up: " 
	<< n_nontree_moved_up << "\n";
    out << "number of total edges moved up: " 
	<< n_tree_moved_up + n_nontree_moved_up << "\n";

    int i;
    out << "Distribution of deleted tree edges at level i (level:#)\n";
    for(i=0; i<=max_level; i++)
	out << i << ":" << n_tree_deletes_at_level[i] << " ";
    out << "\n";
    
    out << "Distribution of deleted nontree edges at level i (level:#)\n";
    for(i=0; i<=max_level; i++)
	out << i << ":" << n_nontree_deletes_at_level[i] << " ";
    out << "\n";
    
#else
    out << "\nhdt_base::print_statistics: sorry, no statistics available\n";
    out << "  compile with -DSTATISTICS to get statistics\n\n";
#endif
}

bool 
hdt_base::connected(node x, node y, int i)
// Return true if x and y are connected on level i. Otherwise
// return false.
{
#ifdef STATISTICS
    n_connected++;
#endif

    // get the active occurrences of x and y at level i
    et_node x_act_occ = (*Gp)[x]->act_occ[i];
    et_node y_act_occ = (*Gp)[y]->act_occ[i];

    // return whether they belong to the same tree at level i
    return (x_act_occ->find_root() == y_act_occ->find_root());
}

void 
hdt_base::insert_tree(edge e, int i, bool create_tree_occs)
// Insert e into F_i. If create_tree_occs is true, then the storage for the
// tree_occ array for e is allocated.
{
#ifdef STATISTICS
    n_ins_tree++;
#endif

    // find the endpoints of e
    node u = source(e);
    node v = target(e);

#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "insert_tree\n"
         << "-----------------------------------------\n"
         << "(" << index(u)<< "," << index(v) << ") tree ins at level "
         << i << "\n"<<flush;
#endif

    // enter level of e
    (*Gp)[e]->level = i;

    // create tree_occ array for e if requested
    if(create_tree_occs)
    {
	(*Gp)[e]->tree_occ = new et_node*[max_level+1];
	for(int lev=max_level; lev>=0; lev--)
	{
	    (*Gp)[e]->tree_occ[lev] = new et_node[4];
	    for(int j=0; j<4; j++) (*Gp)[e]->tree_occ[lev][j] = nil;
	}
    }

    // link the et_trees containing the active occurrences of u and v
    for(int j=i; j>=0; j--) et_link(u,v,e,j,this);
  
    // append e to the list of tree edges at level i
    // edge_info[e]->tree_item = tree_edges[i].append(e);

    // append e to the list of tree edges of u and v
    // update weights of u and v
    (*Gp)[e]->end_point[0] = u;
    (*Gp)[e]->end_point[1] = v;

    (*Gp)[e]->item[0] = (*Gp)[u]->tree_edges[i].append(e);
    (*Gp)[e]->item[1] = (*Gp)[v]->tree_edges[i].append(e);

    if( max_level > 0 )   // otherwise move_tree edges never called
    {
	(*Gp)[u]->act_occ[i]->add_weight(Weight(0,true,false));
	(*Gp)[v]->act_occ[i]->add_weight(Weight(0,true,false));
    }

    // adjust # of components
    components--;
}

void 
hdt_base::delete_tree(edge e)
// Remove the tree edge e from F.
{
#ifdef STATISTICS
    n_del_tree++;
#endif
   
    // get the level of e
    int i = level(e);
   
#ifdef DEBUG
    cout << "-----------------------------------------\n"
	 << "delete_tree of edge"<<flush;
    Gp->print_edge(e);
    cout << "\n-----------------------------------------\n"
	 << "(" << index(source(e)) << "," << index(target(e)) << ") tree del "
         << "at level " << i << "\n"<<flush;
#endif
   
    // cut the spanning trees
    for(int j=i; j>=0; j--) 
	et_cut(e,j,this);
  
    // Find endpoints of e
    node n0 = (*Gp)[e]->end_point[0];
    node n1 = (*Gp)[e]->end_point[1];
  
    // delete e from the list of tree edges of n0 and n1
    (*Gp)[n0]->tree_edges[i].del_item((*Gp)[e]->item[0]);
    (*Gp)[n1]->tree_edges[i].del_item((*Gp)[e]->item[1]);
  
    // set tree_item of e to nil
    (*Gp)[e]->item[0] = nil;
    (*Gp)[e]->item[1] = nil;
    (*Gp)[e]->end_point[0] = nil;
    (*Gp)[e]->end_point[1] = nil;
    
    // update the weights for u and v if necessary 
    if( max_level > 0 )  // otherwise move_tree_edges never called
    {
	if ((*Gp)[n0]->tree_edges[i].empty())
	    (*Gp)[n0]->act_occ[i]->set_weight(Weight(1,false,
		      (*Gp)[n0]->act_occ[i]->get_weight().get_non_tree_bit()));
	if ((*Gp)[n1]->tree_edges[i].empty())
	    (*Gp)[n1]->act_occ[i]->set_weight(Weight(1,false,
                      (*Gp)[n1]->act_occ[i]->get_weight().get_non_tree_bit()));
    }

    // adjust # of components
    components++;
}

void
move_tree_edges_traversal(et_tree root, int i, hdt_base *dc)
{
    list <edge> found_tree_edges;
    list <edge> *curr_list;
    edge e;
    et_node prev, curr;
    bool visited;

    int other;
    node other_node;
    Weight w;

#ifdef DEBUG
    cout << "-----------------------------------------\n"
	 << "move_tree_edges_traversal at level " << i << "\n"
	 << "-----------------------------------------\n"<<flush;
#endif

    prev = nil;
    curr = root;
    visited = false;

    while (curr != nil) 
    {
	if (visited == false) 
	{
	    if (curr->get_subtree_weight().get_tree_bit() == false) 
	    {
		visited = true;
		prev = curr;
		curr = curr->parent();
		continue;
	    }
	    if (curr->get_weight().get_tree_bit() == true) 
	    {
		curr_list = &(*(curr->dc->Gp))[curr->corr_node]->tree_edges[i];

		while (curr_list->size()) 
		{
		    e = curr_list->pop();
		    if ((*(dc->Gp))[e]->end_point[0] == curr->corr_node) 
			other = 1;
		    else other = 0;
		    other_node = (*(dc->Gp))[e]->end_point[other];
		    (*(dc->Gp))[other_node]->tree_edges[i].del_item(
						  (*(dc->Gp))[e]->item[other]);
		    if ((*(dc->Gp))[other_node]->tree_edges[i].empty()) 
		    {
			w = (*(dc->Gp))[other_node]->act_occ[i]->get_weight();
			w.set_tree_bit(false);
			(*(dc->Gp))[other_node]->act_occ[i]->set_weight(w);
		    }
		    found_tree_edges.append(e);
		} // while
	  
                // update bits from path to root
                w = curr->get_weight();
		w.set_tree_bit(false);
		curr->set_weight(w);
	    } // if has tree_edges
        } // end visit

        visited = false;
	if (curr->parent() == prev) 
	{
	    prev = curr;
	    if (curr->left_child()) 
	    { 
		curr = curr->left_child();
		continue;
	    }
	    if (curr->right_child()) 
	    {
		curr = curr->right_child();
		continue;
	    }
	}
	if ((curr->left_child() == prev) && (curr->right_child() != nil)) 
	{
	    prev = curr;
	    curr = curr->right_child();
	    continue;
	} 
	visited = true;
	prev = curr;
	curr = curr->parent();
    }
    // now "insert" edges in i+1
    node n0,n1;
  
    while(found_tree_edges.size()) 
    {
	e = found_tree_edges.pop();
	n0 = source(e);
	n1 = target(e);
    
	// enter level of e
        (*(dc->Gp))[e]->level = i+1;
    
	// update tree_edges[i+1]
        // append e to the list of tree edges of n0 and n1
        // update weights of n0 and n1
    
        (*(dc->Gp))[e]->item[0] = (*(dc->Gp))[n0]->tree_edges[i+1].append(e);
	(*(dc->Gp))[e]->item[1] = (*(dc->Gp))[n1]->tree_edges[i+1].append(e);
      
	(*(dc->Gp))[n0]->act_occ[i+1]->add_weight(Weight(0,true,false));
	(*(dc->Gp))[n1]->act_occ[i+1]->add_weight(Weight(0,true,false));
    
   
	// link the corresponding et_trees at level i+1
        et_link(n0,n1,e,i+1,dc);
    
#ifdef STATISTICS
	dc->n_tree_moved_up++;
#endif
    }
}  


bool
samp_for_replacement(et_tree root, int i, hdt_base *dc, int samp_threshold)
{
    et_node prev, curr;
    bool visited;
    list<edge> curr_list;

    prev = nil;
    curr = root;
    visited = false;

#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "replace:  samp_for_replacement at level" << i <<"\n"
         << "Called on tree rooted at " << flush;
    dc->Gp->print_node(root->corr_node);
    cout << " with threshold " << samp_threshold << "\n"
	 << "-----------------------------------------\n"<<flush;
#endif

    if( root == nil ) 
    {
        cout << "samp_for_replacement: called with nil root\n"<< flush;
	exit(-22);
    }
    
    int no_tree_bits = 0;  // to avoid promoting nontrees 

    while (curr != nil) 
    {
	if (visited == false) 
	{
	    if (curr->get_subtree_weight().get_non_tree_bit() == false) 
	    {
		visited = true;
		prev = curr;
		curr = curr->parent();
		continue;
	    }
	    if( curr->get_weight().get_non_tree_bit() == true )
	    {  
#ifdef DEBUG
		cout << "Found a node " << index(curr->corr_node)
		     <<".\n"
		     << "Scanning list of non-tree edges\n"<<flush;
#endif
		// we have a node with incident non-tree edges of level i 
    	        // check whether any of the edges are replacments
	        // list of non-tree edges incident to node curr
  	        curr_list =
		    (*(curr->dc->Gp))[curr->corr_node]->non_tree_edges[i];
	    
#ifdef DEBUG
		print_list(curr_list,dc->Gp);
#endif
		
		//check each edge of *curr_list to see if it is a replacement.
    	        //if not, increase its level (unless we are in base case)
		edge	f;
		node	u,v;

		while(curr_list.size())
		{
		    f = curr_list.pop();
		    num_sampled++;
#ifdef DEBUG
		    cout << "Non-tree edge: "; curr->dc->Gp->print_edge(f);
		    cout << "\n" <<flush;
#endif
		    u = source(f);
		    v = target(f);

		    if( (*(dc->Gp))[u]->act_occ[i]->find_root() !=
			(*(dc->Gp))[v]->act_occ[i]->find_root() )
		    {   // f is a replacement edge
			dc->delete_non_tree(f);
			dc->insert_tree(f, i, true);
#ifdef DEBUG
			cout << "Found replacement edge "; 
			dc->Gp->print_edge(f);
			cout <<" at level " << i << "\n"
			     << "Finished samp_for_replacement_edge.\n"<<flush;
#endif
			return true;
		    }

		    // heuristic: promote non-replacement edge if there are 
	  	    // no tree edges
		    else if( no_tree_bits )
		    {
			dc->delete_non_tree(f);
			dc->insert_non_tree(f, i+1);
#ifdef DEBUG
			cout << "Moving non-tree edge "<<flush;
			dc->Gp->print_edge(f);
			cout <<" to level " << i+1 << "\n" <<flush;
#endif 
#ifdef STATISTICS
			dc->n_nontree_moved_up++;
#endif
		    }

		    if (num_sampled == samp_threshold) 
			return false;
		} // while
 	    }  // if has non-tree edges
        } // end visit 

        visited = false;
	if (curr->parent() == prev) 
	{
	    prev = curr;
	    if (curr->left_child()) 
	    { 
		curr = curr->left_child();
		continue;
	    }
	    if (curr->right_child()) 
	    {
		curr = curr->right_child();
		continue;
	    }
	}
	if ((curr->left_child() == prev) && (curr->right_child() != nil)) 
	{
	    prev = curr;
	    curr = curr->right_child();
	    continue;
	} 
	visited = true;
	prev = curr;
	curr = curr->parent();
    }
    
#ifdef DEBUG
    cout << "Finished samp_for_replacement_edge.\n"<<flush;
#endif
    return false;
}

bool
find_replacement_edge(et_tree root, int i, hdt_base *dc)
{
    et_node	prev, curr;
    bool	visited;
    list<edge>	curr_list;

    prev = nil;
    curr = root;
    visited = false;

#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "replace:  find_replacement_edge at level" << i <<"\n"
         << "Called on tree rooted at "<<flush;
    dc->Gp->print_node(root->corr_node);
    cout << "\n-----------------------------------------\n"<<flush;
#endif

    if( root == nil ) 
    {
        cout << "find_replacement_edge: called with nil root\n" <<flush;
	exit(-22);
    }
    
   
    while (curr != nil) 
    {
        if (visited == false) 
	{
	    if (curr->get_subtree_weight().get_non_tree_bit() == false) 
	    {
		visited = true;
		prev = curr;
		curr = curr->parent();
		continue;
	    }

	    if( curr->get_weight().get_non_tree_bit() == true )
	    {  
#ifdef DEBUG
	        cout << "Found a node " << index(curr->corr_node)
		     << ".\nScanning List of non-tree edges\n"<<flush;
#endif
		// we have a node with incident non-tree edges of level i 
		// check whether any of the edges are replacments
		// list of non-tree edges incident to node curr
		curr_list = 
		    (*(curr->dc->Gp))[curr->corr_node]->non_tree_edges[i];
	    
#ifdef DEBUG
		print_list(curr_list,dc->Gp);
#endif

		//check each edge of *curr_list to see if it is a replacement.
		//if not, increase its level (unless we are in base case)
		edge	f;
		node	u,v;

		while(curr_list.size())
		{ 
		    f = curr_list.pop();
#ifdef DEBUG
		    cout << "Non-tree edge: "; curr->dc->Gp->print_edge(f);
		    cout << "\n" <<flush;
#endif
		    u = source(f);
		    v = target(f);

		    if( (*(dc->Gp))[u]->act_occ[i]->find_root() !=
			(*(dc->Gp))[v]->act_occ[i]->find_root() )
		    {   
		        // f is a replacement edge
		        dc->delete_non_tree(f);
			dc->insert_tree(f, i, true);
#ifdef DEBUG
			cout << "Found replacement edge "; 
			dc->Gp->print_edge(f);
			cout <<" at level " << i << "\n"
			     << "Finished find_replacement edge.\n"<<flush;
#endif
			return true;
		    } 
		    // Not a replacement edge, so increase level to i+1
		    dc->delete_non_tree(f);
		    dc->insert_non_tree(f, i+1);
#ifdef DEBUG
		    cout << "Moving non-tree edge "; dc->Gp->print_edge(f);
		    cout << " to level " << i+1 << "\n" <<flush;
#endif 
#ifdef STATISTICS
		    dc->n_nontree_moved_up++;
#endif
		} // while list of non-tree edges non-empty
	    } // if has non-tree edges
        } // end visit

	visited = false;
	if (curr->parent() == prev) 
	{
	    prev = curr;
	    if (curr->left_child()) 
	    { 
		curr = curr->left_child();
		continue;
	    }
	    if (curr->right_child()) 
	    {
		curr = curr->right_child();
		continue;
	    }
	}
	if ((curr->left_child() == prev) && (curr->right_child() != nil)) 
	{
	    prev = curr;
	    curr = curr->right_child();
	    continue;
	} 
	visited = true;
	prev = curr;
	curr = curr->parent();
    }
#ifdef DEBUG
    cout << "Finished find_replacement_edge.\n" <<flush;
#endif
  return false;
}  // find_replacement_edge


void 
hdt_base::replace(node v, node w, int i)
// try to reconnect the trees on level i containing v and w.
// if not possible try to recurse on lower level
{
/*
	Find T_v and T_w (in F_i), the trees that contain v and w.
	Increase all edges of T_v to level i+1.
	Visit level i non-tree edges f incident to T_v.
	   if f connects T_v and T_w (if f's other endpoint is in T_w)
		insert f into F_i
		return
	   else
		level(f)++
*/

#ifdef STATISTICS
    cout << n_replace << "\t" << n_nontree_moved_up << "\n" << flush;
    n_replace++;
#endif

#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "replace at level " << i << "\n"
         << "-----------------------------------------\n" <<flush;
#endif

    num_sampled = 0;
    for( int j=i; j>=0; j-- )
    {
	// determine the level j trees containing v and w
        et_tree Tv = (*Gp)[v]->act_occ[j]->find_root();
	et_tree Tw = (*Gp)[w]->act_occ[j]->find_root();

#ifdef DEBUG
	cout << "Roots are " << index(v) << " and " << index(w) << ".\n"
	     << "Tree sizes are " << (Tv->get_subtree_weight()).get_size() 
	     << " and " << (Tw->get_subtree_weight()).get_size() << ".\n";
#endif

	// let Tv be the smaller tree
        if(Tv->get_subtree_weight() > Tw->get_subtree_weight()) 
	    Tv = Tw;

	int samp_threshold = -1;
	if (SAMP_THRESHOLD_EDGES > 0) 
	{
	    samp_threshold = SAMP_THRESHOLD_EDGES;
#ifdef DEBUG
	    cout << "Sample threshold kicks in\n"; 
#endif DEBUG
	}

	if (BASE_SIZE > 0) {
	    if (Tv->get_subtree_weight() < BASE_SIZE) {
		samp_threshold = INFINITY;
#ifdef DEBUG
		cout << "Base case kicks in\n"; 
#endif DEBUG
	    }
	}

	if ((samp_threshold > 0) &&
	    (samp_for_replacement(Tv, j, this, samp_threshold)) ) {
#ifdef STATISTICS
	    n_replace_found++;
#endif
	    return;
	}

	// At this point samp_for_replacement has failed
        // if in base case, simply return
        // otherwise promote tree edges and do normal find_replace search
	if (num_sampled < samp_threshold)
	    continue;

	move_tree_edges_traversal(Tv, j, this);
	if (find_replacement_edge(Tv, j, this)) {
#ifdef STATISTICS
	    n_replace_found++;
#endif STATISTICS
	    return;
	}
    }
}

void 
hdt_base::insert_non_tree(edge e, int i)
// Insert the non-tree edge e into G_i.
{
#ifdef STATISTICS
    n_ins_non_tree++;
#endif
	
#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "insert_nontree\n"
         << "-----------------------------------------\n"
         << "(" << index(source(e)) << "," << index(target(e))
         << ") non-tree ins at level " << i << "\n" <<flush;
#endif
	
    (*Gp)[e]->level = i;
    node u = source(e);
    node v = target(e);
	
    // insert e in the item lists of its endpoints at level i
    (*Gp)[e]->end_point[0] = u;
    (*Gp)[e]->end_point[1] = v;

    (*Gp)[e]->item[0] = (*Gp)[u]->non_tree_edges[i].append(e);
    (*Gp)[e]->item[1] = (*Gp)[v]->non_tree_edges[i].append(e);
	
    // increase the weight of the active occurrences of u and v at level i
    (*Gp)[u]->act_occ[i]->add_weight(Weight(0,false,true));
    (*Gp)[v]->act_occ[i]->add_weight(Weight(0,false,true));
}

void 
hdt_base::delete_non_tree(edge e)
// Delete the non-tree edge e.
{
#ifdef STATISTICS
    n_del_non_tree++;
#endif

    // find the endpoints and the level of e
    // node u = source(e);
    // node v = target(e);
    int i = level(e);

#ifdef DEBUG
    cout << "-----------------------------------------\n"
         << "delete_nontree\n"
         << "-----------------------------------------\n"
         << "(" << index(source(e)) << "," << index(target(e))
         << ") non-tree del at level " << i << "\n" <<flush;
#endif

    node n0 = (*Gp)[e]->end_point[0];
    node n1 = (*Gp)[e]->end_point[1];

    // remove e from the non_tree lists of u and v at level i
    (*Gp)[n0]->non_tree_edges[i].del_item((*Gp)[e]->item[0]);
    (*Gp)[n1]->non_tree_edges[i].del_item((*Gp)[e]->item[1]);

    (*Gp)[e]->item[0] = nil;
    (*Gp)[e]->item[1] = nil;
    (*Gp)[e]->end_point[0] = nil;
    (*Gp)[e]->end_point[1] = nil;
    
    // update the weights for n0 and n1 if necessary
    if ((*Gp)[n0]->non_tree_edges[i].empty())
	(*Gp)[n0]->act_occ[i]->set_weight(
	   Weight(1,(*Gp)[n0]->act_occ[i]->get_weight().get_tree_bit(),false));
    if ((*Gp)[n1]->non_tree_edges[i].empty())
	(*Gp)[n1]->act_occ[i]->set_weight(
           Weight(1,(*Gp)[n1]->act_occ[i]->get_weight().get_tree_bit(),false));
}


hdt_base::hdt_base(dc_graph& G, int n_levels, int samp_threshold_edges, 
		   int base_size)
// initializes the dynamic connectivity data structure
// if n_levels > 0 then it specifies the number of levels (default O(log n))
//
//  Note that the indexing of levels by HDT is the opposite of 
//  that used by Henzinger and King; in HDT, F_i contains 
//  F_{i+1}.
{
    // --- initialize the constants ---
    Gp = &G;
    int log_n = 0;
    int i;
    for(i = G.number_of_nodes(); i; i /= 2) 
	log_n++;

    if(n_levels > 0) 
	max_level = n_levels - 1;
    else 
	max_level = log_n;

    SAMP_THRESHOLD_EDGES = samp_threshold_edges;
    BASE_SIZE = base_size;

    if (BASE_SIZE > 0) {
	int log_base_size = 0;
	if( BASE_SIZE > G.number_of_nodes() )
	    max_level = 0;
	else if( BASE_SIZE >= 0 )
	{
	    for(i = BASE_SIZE; i; i /= 2) log_base_size++;
	    max_level = log_n - log_base_size;
	}
	else
	{
	    cout << "error: hdt_base (constructor): BASE_SIZE= "
		 << BASE_SIZE << " < 0\n" << flush;
	    exit( -117 );
	}
    }
    
    components = G.number_of_nodes();
  
#ifdef DEBUG
    cout << "|V(G)|          = " << G.number_of_nodes() << "\n"
	 << "MAX_LEVEL       = " <<  max_level << "\n\n"<<flush;
#endif
  
    // --- initialize dummy nodes ---
    et_dummy = new et_node_struct(this,nil);
  
    // --- initialize the nodes ---
    node u;
    forall_nodes(u,G)
    {
	G[u] = new hdt_node_struct();
	G[u]->act_occ = new et_node [max_level+1];
	G[u]->tree_edges = new list<edge> [max_level+1];
	G[u]->non_tree_edges = new list<edge> [max_level+1];

	for(i=0; i<=max_level; i++)
	    G[u]->act_occ[i] = new et_node_struct(this,u,i,true);
    }

    // --- initialize the edges ---
    edge e;
    forall_edges(e,G)
    {
	G[e] = new hdt_edge_struct();
	if(!connected(source(e),target(e),0)) insert_tree(e,0,true);
	else                                  insert_non_tree(e,0);
    }
    
#ifdef STATISTICS
    // --- initialize statistics ---
    n_ins = 0;
    n_del = 0;
    n_query = 0;
    n_connected = 0;
    n_ins_tree = 0;
    n_del_tree = 0;
    n_replace = 0;
    n_ins_non_tree = 0;
    n_del_non_tree = 0;
    n_replace_found = 0;
    n_tree_moved_up = 0;
    n_nontree_moved_up = 0;

    n_nontree_deletes_at_level = new unsigned int[max_level+1];
    n_tree_deletes_at_level = new unsigned int[max_level+1];
    for(i=0; i<=max_level; i++) 
    {
	n_nontree_deletes_at_level[i] = 0;
	n_tree_deletes_at_level[i] = 0;
    }
#endif
}

//void 
hdt_base::~hdt_base()
{
    // first delete all edges in the data structure (not in G)
    edge e;
    forall_edges(e,*Gp)
    {
	if(tree_edge(e))
	{
	    delete_tree(e);
	    for(int j=0; j<=max_level; j++) delete[] (*Gp)[e]->tree_occ[j];
	    delete[] (*Gp)[e]->tree_occ;
	}
	else delete_non_tree(e);
	
	delete (*Gp)[e];
	(*Gp)[e] = nil;
    }

#ifdef STATISTICS
    delete[] n_nontree_deletes_at_level;
    delete[] n_tree_deletes_at_level;
#endif

    // delete the et_nodes and the information at the nodes of G
    node v;
    forall_nodes(v,*Gp)
    {
	// per node of G only its active occurrence at each level is left
        for(int i=0; i<=max_level; i++) delete (*Gp)[v]->act_occ[i];

	delete[] (*Gp)[v]->act_occ;
	delete[] (*Gp)[v]->tree_edges;
	delete[] (*Gp)[v]->non_tree_edges;

	delete (*Gp)[v];
	(*Gp)[v] = nil;
    }

    // delete dummy nodes
    delete et_dummy;
}

// ------------------------------------------------------------- //
// auxiliary data structures                                     //
// ------------------------------------------------------------- //

// ------------------------------------------------------------- //
// et_tree.c: implementation of et_trees.                        //
//                                                               //
// Version 1.3                                                   //
//                                                               //
// Copyright (C) 1995 by David Alberts                           //
// FU Berlin, Inst. f. Informatik, Takustr. 9, 14195 Berlin, FRG //
// All rights reserved.                                          //
// There is ABSOLUTELY NO WARRANTY.                              //
// ------------------------------------------------------------- //

// declarations in "dc_hdt.h" 

et_node_struct::et_node_struct
(hdt_base* dcp, node v, int my_level, int activate)
: hdt_rnbw_node_struct(Weight(1))
{
  dc = dcp;
  corr_node = v;
  level = my_level;
  active = activate;
  if(activate) dc->Gp->inf(v)->act_occ[level] = this;
  edge_occ[0] = edge_occ[1] = nil;
}

et_node_struct::et_node_struct(et_node en)
: hdt_rnbw_node_struct(0)
{
  dc = en->dc;
  corr_node = en->corr_node;
  level = en->level;
  active = false;
  edge_occ[0] = edge_occ[1] = nil;
}

void 
et_node_struct::pass_activity(et_node to)
// Make this node inactive and pass its activity to to.
// Prec.: this node is active, to represents the same vertex and is on the
//        same level.
{
  active = false;
  to->active = true;
  to->set_weight(get_weight());
  set_weight(Weight(0,false,false));
  dc->Gp->inf(corr_node)->act_occ[level] = to;
}

void 
change_root(et_tree& et, et_node en, int i, hdt_base* dc)
// Change the root of the tree T represented by the et_tree et to the
// vertex represented by the et_node en. The new tree is stored at et.
// Prec.: The et_node en is in the et_tree et.
{
  et_node first_nd = et->first();

  // if en is already the first node do nothing
  if(en == first_nd) return;

  // create a new occurrence for the new root
  et_node new_occ = new et_node_struct(en);

  // --- update active occurrences --- //
  et_node last_nd = et->last();
  // if the first node is active, pass activity to last node, since
  // the first node will be deleted
  if(first_nd->active) first_nd->pass_activity(last_nd);

  // --- update tree_occs --- //
  if(en->edge_occ[rnb_left] == en->edge_occ[rnb_right])
  {
    // replace the nil pointer in tree_occs of en->edge_occ[rnb_left]
    // by the new occurrence
    int k;
    for(k=0; nil != dc->Gp->inf(en->edge_occ[rnb_left])->tree_occ[i][k];
        k++);
    dc->Gp->inf(en->edge_occ[rnb_left])->tree_occ[i][k] = new_occ;
  }
  else
  {
    // replace en by the new occurrence
    int k;
    for(k=0; en != dc->Gp->inf(en->edge_occ[rnb_left])->tree_occ[i][k];
        k++);
    dc->Gp->inf(en->edge_occ[rnb_left])->tree_occ[i][k] = new_occ;
  }

  edge first_edge = first_nd->edge_occ[rnb_right];
  if((first_edge != last_nd->edge_occ[rnb_left]) ||
     (en == last_nd))
  {
    // replace first_nd by last_nd in the tree_occs of first_edge
    int k;
    for(k=0; first_nd != dc->Gp->inf(first_edge)->tree_occ[i][k];
        k++);
    dc->Gp->inf(first_edge)->tree_occ[i][k] = last_nd;
  }
  else
  {
    // replace first_nd by nil in the tree_occs of first_edge
    int k;
    for(k=0; first_nd != dc->Gp->inf(first_edge)->tree_occ[i][k];
        k++);
    dc->Gp->inf(first_edge)->tree_occ[i][k] = nil;
  }

  // --- update edge_occs --- //
  // right edge of first_nd becomes right edge of last node
  last_nd->edge_occ[rnb_right] = first_edge;

  // left edge of en becomes left edge of new_occ
  new_occ->edge_occ[rnb_left] = en->edge_occ[rnb_left];
  en->edge_occ[rnb_left] = nil;

  // --- update the et_tree --- //
  // split off first occurrence and delete it
  rnb_tree s1, s2;
  split(first_nd,rnb_right,s1,s2,dc->et_dummy);
  delete first_nd;

  // split immediately before en
  split(en,rnb_left,s1,s2,dc->et_dummy);

  // join the pieces
  et = et_join((et_tree)s2,et_join((et_tree)s1,new_occ,dc->et_dummy),
                dc->et_dummy);
}

et_tree 
et_link(node u, node v, edge e, int i, hdt_base* dc)
// Modify the et_trees at level i corresponding to the insertion of
// the edge e = (u,v) into F_i.
// Prec.: u and v belong to dc, and they are not connected at the
//        valid level i.
{
  // get active occurrences of u and v, create a new occurrence of u
  et_node u_act = dc->Gp->inf(u)->act_occ[i];
  et_node v_act = dc->Gp->inf(v)->act_occ[i];
  et_node new_u_occ = new et_node_struct(u_act);

  // find the tree etv containing v_act and reroot it at v_act
  et_tree etv = v_act->find_root();
  change_root(etv,v_act,i,dc);

  // --- initialize tree_occs of e ---
  // u_act and new_u_occ become the first two tree_occs of e
  dc->Gp->inf(e)->tree_occ[i][0] = u_act;
  dc->Gp->inf(e)->tree_occ[i][1] = new_u_occ;

  // the first and the last node of etv are tree_occ[i][2 and 3] if
  // they are different otherwise tree_occ[i][2] is nil
  et_node etv_last = etv->last();
  dc->Gp->inf(e)->tree_occ[i][3] = etv_last;
  if(etv_last != v_act) dc->Gp->inf(e)->tree_occ[i][2] = v_act;
  else                  dc->Gp->inf(e)->tree_occ[i][2] = nil;

  // --- update tree_occs of the edge following e if it exists ---
  edge after_e = u_act->edge_occ[rnb_right];
  if(after_e)
  {
    if(u_act->edge_occ[rnb_left] != after_e)
    {
      // replace u_act by new_u_occ
      int k;
      for(k=0; u_act != dc->Gp->inf(after_e)->tree_occ[i][k]; k++);
      dc->Gp->inf(after_e)->tree_occ[i][k] = new_u_occ;
    }
    else
    {
      // replace nil pointer by new_u_occ
      int k;
      for(k=0; nil != dc->Gp->inf(after_e)->tree_occ[i][k]; k++);
      dc->Gp->inf(after_e)->tree_occ[i][k] = new_u_occ;
    }
  }

  // --- update edge_occs ---
  new_u_occ->edge_occ[rnb_right] = u_act->edge_occ[rnb_right];
  new_u_occ->edge_occ[rnb_left] = e;
  u_act->edge_occ[rnb_right] = e;
  v_act->edge_occ[rnb_left] = e;
  etv_last->edge_occ[rnb_right] = e;

  // --- update et_trees ---
  // concatenate etv and the new occurrence
  etv = et_join(etv,new_u_occ,dc->et_dummy);

  // split the et_tree containing u_act after u_act
  rnb_tree s1, s2;
  split(u_act,rnb_right,s1,s2,dc->et_dummy);

  // concatenate the pieces
  return et_join((et_tree)s1,et_join(etv,(et_tree)s2,dc->et_dummy),
                  dc->et_dummy);
}

void 
et_cut(edge e, int i, hdt_base* dc)
// Update the et_trees at level corresponding to the removal of
// the tree edge e.
// Prec.: e actually is a tree edge at level i.
{
  // get the et_nodes representing e at level i
  et_node ea1 = dc->Gp->inf(e)->tree_occ[i][0];
  et_node ea2 = dc->Gp->inf(e)->tree_occ[i][1];
  et_node eb1 = dc->Gp->inf(e)->tree_occ[i][2];
  et_node eb2 = dc->Gp->inf(e)->tree_occ[i][3];

  // set the tree_occ to nil;
  dc->Gp->inf(e)->tree_occ[i][0] = nil;
  dc->Gp->inf(e)->tree_occ[i][1] = nil;
  dc->Gp->inf(e)->tree_occ[i][2] = nil;
  dc->Gp->inf(e)->tree_occ[i][3] = nil;

  // sort ea1, ea2, eb1, and eb2, such that
  // ea1 < eb1 < eb2 < ea2 in In-order if they are not nil
  // eb1 may be nil
  et_node aux;
  if(ea1 && ea2)
  {
    if(smaller(ea2,ea1))
    {
      aux = ea1; ea1 = ea2; ea2 = aux;
    }
  }
  else      // either ea1 or ea2 is nil...
  {
    if(ea1) // ...it is ea2
    {
      ea2 = ea1; ea1 = nil;
    }
  }

  if(eb1 && eb2)
  {
    if(smaller(eb2,eb1))
    {
      aux = eb1; eb1 = eb2; eb2 = aux;
    }
  }
  else      // either eb1 or eb2 is nil...
  {
    if(eb1) // ...it is eb2
    {
      eb2 = eb1; eb1 = nil;
    }
  }

  // now ea2 and eb2 are non-nil
  if(smaller(ea2,eb2))
  {
    aux = eb1; eb1 = ea1; ea1 = aux;
    aux = eb2; eb2 = ea2; ea2 = aux;
  }

  // --- update et_trees ---
  // compute s1, s2 and s3
  rnb_tree s1, s2, s3;
  split(ea1,rnb_right,s1,s2,dc->et_dummy);
  split(ea2,rnb_right,s2,s3,dc->et_dummy);

  // compute the first of the two resulting trees
  et_join((et_tree)s1,(et_tree)s3,dc->et_dummy);

  // split off ea2 from s2 giving the second tree
  split(eb2,rnb_right,s1,s2,dc->et_dummy);

  // --- update active occurrences ---
  if(ea2->active) ea2->pass_activity(ea1);

  // --- update tree_occs ---
  // update tree_occs of the edge following e if it exists
  edge after_e = ea2->edge_occ[rnb_right];
  if(after_e)
  {
    if(ea1->edge_occ[rnb_left] != after_e)
    {
      // replace ea2 by ea1
      int k;
      for(k=0; ea2 != dc->Gp->inf(after_e)->tree_occ[i][k]; k++);
      dc->Gp->inf(after_e)->tree_occ[i][k] = ea1;
    }
    else
    {
      // replace ea2 by nil
      int k;
      for(k=0; ea2 != dc->Gp->inf(after_e)->tree_occ[i][k]; k++);
      dc->Gp->inf(after_e)->tree_occ[i][k] = nil;
    }
  }

  // --- update edge_occs --- //
  ea1->edge_occ[rnb_right] = ea2->edge_occ[rnb_right];
  if(eb1) eb1->edge_occ[rnb_left] = nil;
  else    eb2->edge_occ[rnb_left] = nil;
  eb2->edge_occ[rnb_right] = nil;

  delete ea2;
}

void 
et_node_struct::print()
// we redefine print in order to output also the additional fields
{
  // output base fields
  hdt_rnbw_node_struct::print();

  // output new fields
  cout << "  dc:          " << dc << "\n";cout<<flush;

  cout << "  corr_node:   " << index(corr_node) << "\n";
  cout << "  level:       " << level << "\n";
  cout << "  active:      " << active << "\n";cout<<flush;

  if(edge_occ[0])
  {
    cout << "  edge_occ[0]: " << "(" << index(source(edge_occ[0]));
    cout << "," << index(target(edge_occ[0])) << ")\n";
  }
  else cout << "  edge_occ[0]: nil\n";
  if(edge_occ[1])
  {
    cout << "  edge_occ[1]: " << "(" << index(source(edge_occ[1]));
    cout << "," << index(target(edge_occ[1])) << ")\n";
  }
  else cout << "  edge_occ[1]: nil\n";cout<<flush;

}

// ------------------------------------------------------------- //
// hdt_rnbw_tree.c: implements rnbw_trees for HDT                //
//                                                               //
// comment: hdt_rnbw_trees are a modified version of David       //
//	    Alberts' rnbw_trees (derived from rnb_trees).	 //
//	    They have a non-negative weight and a bitmap	 //
//	    at each node and subtree weights.                    //
// ------------------------------------------------------------- //

// declarations in "hdt_rnbw_tree.h"

void 
hdt_rnbw_node_struct::after_rot()
{
  // the parent gets the sub_weight of this node
  parent()->sub_weight = sub_weight;

  // recalculate the sub_weight field of this node
  sub_weight = weight;
  if(left_child()) sub_weight += left_child()->sub_weight;
  if(right_child()) sub_weight += right_child()->sub_weight;
}

void 
hdt_rnbw_node_struct::init()
{
  // initialize the sub_weight field of this node
  sub_weight = weight;
  if(left_child()) sub_weight += left_child()->sub_weight;
  if(right_child()) sub_weight += right_child()->sub_weight;
}

void 
hdt_rnbw_node_struct::isolate()
// We fix the sub_weight fields of the ancestors of this node.
{
    hdt_rnbw_node aux = parent();
    if (aux)
    {
        int dir = (aux->left_child() == this) ? rnb_right : rnb_left;
	aux->sub_weight = aux->weight;
	if (aux->child[dir]) aux->sub_weight += 
			       ((hdt_rnbw_node)aux->child[dir])->sub_weight;

	for(aux = aux->parent() ; aux; aux = aux->parent())
	{
	    aux->sub_weight = aux->weight;
	    if (aux->left_child()) 
	        aux->sub_weight += aux->left_child()->sub_weight;
	    if (aux->right_child()) 
	      aux->sub_weight += aux->right_child()->sub_weight;
	}
    }
	
    // fix base class
    rnb_node_struct::isolate();
}

void 
hdt_rnbw_node_struct::set_weight(Weight w)
// sets the weight of this node to w
// Prec.: w >= 0
{
    // remember the difference between the new and the old weight
    // int w_diff = w.get_size() - weight.get_size();
	
    // update the weight field of this node
    weight = w;
    
    // update the subweight field of myself and ancestors
    for( hdt_rnbw_node aux = this; aux; aux = aux->parent())
    {
        aux->sub_weight = aux->weight;
	if (aux->left_child()) 
	    aux->sub_weight += aux->left_child()->sub_weight;
	if (aux->right_child()) 
	    aux->sub_weight += aux->right_child()->sub_weight;
    }
}

void 
hdt_rnbw_node_struct::add_weight(Weight w)
// adds w to the weight of this node
// Prec.: w >= -(weight of this node)
{
  // update the weight and subweight fields of this node
  sub_weight += w;
  weight += w;

  // update the sub_weight fields of the ancestors of this node
  for(hdt_rnbw_node aux = parent(); aux; aux = aux->parent())
    aux->sub_weight += w;
}

void 
hdt_rnbw_node_struct::print()
// we redefine print in order to output also the additional fields
{
  // output base fields
  rnb_node_struct::print();

  // output new fields
  cout << "  weight:      " << weight << "\n";cout<<flush;
  cout << "  sub_weight:  " << sub_weight << "\n";cout<<flush;

}


// ------------------------------------------------------------- //
// rnb_tree.c: implementation of rnb_trees                       //
//                                                               //
// comment: rnb_tree is an implementation of balanced binary     //
//          trees with a randomized balancing scheme.            //
//          See also the documentation in dyn_con.ps.            //
//                                                               //
// Version 1.3                                                   //
//                                                               //
// Copyright (C) 1995 by David Alberts                           //
// FU Berlin, Inst. f. Informatik, Takustr. 9, 14195 Berlin, FRG //
// All rights reserved.                                          //
// There is ABSOLUTELY NO WARRANTY.                              //
// ------------------------------------------------------------- //

// declarations in "rnb_tree.h"

random_source rnb_node_struct::rs;    // define the static member

inline void 
rotate(rnb_node  rot_child, rnb_node rot_parent)
// Rotate such that rot_child becomes the parent of rot_parent.
// Prec.: rot_child is a child of rot_parent.
{
  // determine the direction dir of the rotation
  int dir = (rot_parent->child[rnb_left] == rot_child) ? rnb_right : rnb_left;

   // subtree which changes sides
  rnb_tree middle = rot_child->child[dir];

  // fix middle tree
  rot_parent->child[1-dir] = middle;
  if(middle) middle->par = rot_parent;

  // fix parent field of rot_child
  rot_child->par = rot_parent->par;
  if(rot_child->par)
    if(rot_child->par->child[rnb_left] == rot_parent)
       rot_child->par->child[rnb_left]  = rot_child;
    else
       rot_child->par->child[rnb_right] = rot_child;

  // fix parent field of rot_parent
  rot_child->child[dir] = rot_parent;
  rot_parent->par = rot_child;

  // fix additional information in derived classes
  rot_parent->after_rot();
}

void 
rnb_node_struct::isolate()
// Make this node an isolated node.
// Prec.: this != nil
{
  // adjust child pointer of parent if it exists
  if(par)
    if(par->child[rnb_left] == this)
      par->child[rnb_left] = nil;
    else
      par->child[rnb_right] = nil;

  // adjust parent pointers of children if they exist
  if(child[rnb_left]) child[rnb_left]->par = nil;
  if(child[rnb_right]) child[rnb_right]->par = nil;

  par = child[rnb_left] = child[rnb_right] = nil;
}

rnb_tree 
rnb_node_struct::find_root()
// returns the root of the tree containing this node.
// Prec.: this != nil
{
  rnb_node aux;
  for(aux = this; aux->par; aux = aux->par);
  return aux;
}

rnb_node 
rnb_node_struct::sub_pred()
// returns the predecessor of this node in the subtree rooted at this node
// or nil if it does not exist
// Prec.: this != nil
{
  // handle the nil case first
  if(!child[rnb_left]) return nil;

  // find the last node with no right child in the left subtree of u
  rnb_node aux;
  for(aux = child[rnb_left]; aux->child[rnb_right];
      aux = aux->child[rnb_right]);
  return aux;
}

rnb_node 
rnb_node_struct::sub_succ()
// returns the successor of this node in the subtree rooted at this node
// or nil if it does not exist
// Prec.: this != nil
{
  // handle the nil case first
  if(!child[rnb_right]) return nil;

  // find the first node with no left child in the right subtree of u
  rnb_node aux;
  for(aux = child[rnb_right]; aux->child[rnb_left];
      aux = aux->child[rnb_left]);
  return aux;
}

rnb_node 
rnb_node_struct::pred()
// returns the predecessor of this node or nil if it does not exist
// Prec.: this != nil
{
  // search for predecessor in the subtree of this node first
  rnb_node sub_pr = sub_pred();
  // if it exists we can return it
  if(sub_pr) return sub_pr;

  // otherwise we have to look for the ancestors of this node
  if(par)       // if there is a parent
    if(this == par->child[rnb_right])
    // this is a right child
      return par;
    else
    // this is a left child
    {
      for(rnb_node aux = par; aux->par; aux = aux->par)
        if(aux == aux->par->child[rnb_right]) return aux->par;
    }

  // there is no predecessor
  return nil;
}

rnb_node 
rnb_node_struct::succ()
// returns the successor of this node or nil if it does not exist
// Prec.: this != nil
{
  // search for successor in the subtree of this node first
  rnb_node sub_s = sub_succ();
  // if it exists we can return it
  if(sub_s) return sub_s;

  // otherwise we have to look for the ancestors of this node
  if(par)       // if there is a parent of u
    if(this == par->child[rnb_left])
    // this node is a left child
      return par;
    else
    // this node is a right child
    {
      for(rnb_node aux = par; aux->par; aux = aux->par)
        if(aux == aux->par->child[rnb_left]) return aux->par;
    }

  // there is no predecessor
  return nil;
}

rnb_node 
rnb_node_struct::first()
// Return the first node in In-order in the tree rooted at this node.
{
  // remember one node before current node
  rnb_node last = nil;
  for(rnb_node current = this; current; current = current->child[rnb_left])
    last = current;

  return last;
}

rnb_node 
rnb_node_struct::last()
// Return the last node of this tree.
{
  // remember one node before current node
  rnb_node last = nil;
  for(rnb_node current = this; current; current = current->child[rnb_right])
    last = current;

  return last;
}

int 
smaller(rnb_node u, rnb_node v)
// returns true iff u is smaller than v
{
  if(!u || !v) return false;
  if(u == v) return false;

  // determine the height of u and v
  rnb_node aux_u = u;
  int u_height, v_height;
  for(u_height = 0; aux_u->par; aux_u = aux_u->par, u_height++);
  rnb_node aux_v = v;
  for(v_height = 0; aux_v->par; aux_v = aux_v->par, v_height++);

  // if u and v have different roots they are incomparable and we return false
  if(aux_u != aux_v) return false;

  // we represent the paths from u and v to their roots by arrays
  // create arrays
  int *u_path = new int[u_height];
  int *v_path = new int[v_height];

  // insert left and right moves
  int u_i = u_height - 1;
  for(aux_u = u; aux_u->par; aux_u = aux_u->par, u_i--)
  {
    if(aux_u->par->child[rnb_left] == aux_u) u_path[u_i] = rnb_left;
    else                                     u_path[u_i] = rnb_right;
  }

  int v_i = v_height - 1;
  for(aux_v = v; aux_v->par; aux_v = aux_v->par, v_i--)
  {
    if(aux_v->par->child[rnb_left] == aux_v) v_path[v_i] = rnb_left;
    else                                     v_path[v_i] = rnb_right;
  }

  // compare the paths
  // skip identical prefix
  int i;
  for(i = 0; ((i<u_height) && (i<v_height)) && (u_path[i] == v_path[i]);
      i++);

  // at least one path is not completely scanned because u!=v
  // but u->find_root() == v->find_root()
  // at i they are different
  int result;
  if( (i<u_height) && (u_path[i] == rnb_left) )
    result = true;
  else
    if( (i<v_height) && (v_path[i] == rnb_right) )
      result = true;
    else
      result = false;

  // delete the paths
  delete[] u_path;
  delete[] v_path;

  return result;
}

rnb_tree 
rnb_join(rnb_tree t1, rnb_tree t2, rnb_node dummy)
// join t1 and t2 and return the resulting rnb_tree
{
  // handle the trivial t1 == nil || t2 == nil case
  if(!t1 || !t2)
  {
    if(t1) return t1;
    if(t2) return t2;
    return nil;
  }

  dummy->par = nil;
  dummy->child[rnb_left] = t1;
  dummy->child[rnb_right] = t2;
  t1->par = dummy;
  t2->par = dummy;
  // fix additional information in derived classes
  dummy->init();

  // trickle dummy down
  while( (dummy->child[rnb_left]) || (dummy->child[rnb_right]) )
  // while there is at least one child
  {
    // rotate with child with biggest priority

    // find child with biggest priority...
    rnb_node bigger = dummy->child[rnb_left];
    if(dummy->child[rnb_right])
    {
      if(dummy->child[rnb_left])
      {
        if(dummy->child[rnb_right]->prio > dummy->child[rnb_left]->prio)
          bigger = dummy->child[rnb_right];
      }
      else bigger = dummy->child[rnb_right];
    }

    // ...and rotate with it
    rotate(bigger,dummy);
  }

  // disconnect dummy from the new tree
  dummy->isolate();

  // return root of the new tree
  if(t2->par) return t1;
  else        return t2;
}

void 
split(rnb_node at, int where, rnb_tree& t1, rnb_tree& t2, rnb_node dummy)
// split the rnb_tree containing the node at before or after at
// depending on where. If where == rnb_left we split before at,
// else we split after at. The resulting trees are stored in t1 and t2.
// If at == nil, we store nil in t1 and t2.
{
  // handle the trivial at == nil case first
  if(!at)
  {
    t1 = nil;
    t2 = nil;
    return;
  }

  dummy->child[rnb_left] = nil;
  dummy->child[rnb_right] = nil;

  // insert dummy in the right place (w.r.t. In-order)
  // where == rnb_left => split before at
  // where != rnb_left => split after at
  if(where != rnb_left) // split after at
  {
    // store dummy as left child of the subtree successor of at
    // or as right child of at if there is no subtree successor
    rnb_node s = at->sub_succ();
    if(!s)
    {
      at->child[rnb_right] = dummy;
      dummy->par = at;
    }
    else
    {
      s->child[rnb_left] = dummy;
      dummy->par = s;
    }
  }
  else     // split before at
  {
    // store dummy as right child of the subtree predecessor of at
    // or as left child of at if there is no subtree predecessor
    rnb_node p = at->sub_pred();
    if(!p)
    {
      at->child[rnb_left] = dummy;
      dummy->par = at;
    }
    else
    {
      p->child[rnb_right] = dummy;
      dummy->par = p;
    }
  }
  // fix additional information in derived classes
  dummy->init();


  // rotate dummy up until it becomes the root
  for(rnb_node u = dummy->par; u; u = dummy->par) rotate(dummy,u);

  // store the subtrees of dummy in t1 and t2
  t1 = dummy->child[rnb_left];
  t2 = dummy->child[rnb_right];

  // disconnect dummy
  dummy->isolate();
}

void 
rnb_node_struct::print()
// prints the contents of this node to stdout for testing
// Prec.: this != nil
{
  cout << "node at " << this << ":\n";cout<<flush;

  cout << "  parent:      " << par << "\n";cout<<flush;

  cout << "  left child:  " << child[rnb_left] << "\n";cout<<flush;

  cout << "  right child: " << child[rnb_right] << "\n";cout<<flush;

  cout << "  priority:    " << prio << "\n";cout<<flush;

}

void 
traverse(rnb_tree t)
// traverses the tree and outputs each node to stdout for testing
{
  if(t)
  {
    t->print();
    traverse(t->child[rnb_left]);
    traverse(t->child[rnb_right]);
  }
}

