// ------------------------------------------------------------- //
// dyn_con.c: implementation of internal functions and user      //
//            interface functions for the implementation of      //
//            the dynamic connectivity algorithm by              //
//            M. Rauch Henzinger and V. King.                    //
//                                                               //
//            See also the documentation in dyn_con.ps.          //
//                                                               //
// Version 1.3                                                   //
//                                                               //
// Copyright (C) 1995 by David Alberts                           //
// FU Berlin, Inst. f. Informatik, Takustr. 9, 14195 Berlin, FRG //
// All rights reserved.                                          //
// There is ABSOLUTELY NO WARRANTY.                              //
// ------------------------------------------------------------- //

// RCS Id //
static char rcs[]="$Id: dyn_con.fw,v 1.17 1995/11/22 10:13:55 alberts Exp $";

#include"dyn_con.h"
#include<LEDA/queue.h>
#include<sys/time.h>

edge dyn_con::ins(node u, node v)
// create an edge connecting u and v and return it
{
#ifdef STATISTICS
  n_ins++;
#endif
  // create the new edge
  edge e = Gp->new_edge(u,v);
  (*Gp)[e] = new dc_edge_struct();

  // test whether u and v are already connected
  if(!connected(u,v,max_level))
  // they are not, so e becomes a forest edge at level max_level
  {
    insert_tree(e,max_level,true);
    added_edges[max_level]++;
    rebuild(max_level);
  }
  else
  // u and v are already connected, find lowest such level
  {
    // current level
    int curr_level = max_level/2;
    // lower bound and upper bound
    int lower = 0;
    int upper = max_level;
    while(curr_level != lower)
    {
      if(connected(u,v,curr_level))
      // search below current level
      {
        upper = curr_level;
        curr_level = (lower + curr_level)/2;
      }
      else
      // search above current level
      {
        lower = curr_level;
        curr_level = (upper + curr_level)/2;
      }
    }
    // Now depending on parity either
    // a) connected(u,v,lower-1) == false && connected(u,v,lower) == true or
    // b) connected(u,v,lower) == false && connected(u,v,lower+1) == true holds.

    // handle case a),
    if(!connected(u,v,lower)) lower++;

    // insert e at appropriate level,
    insert_non_tree(e,lower);
    added_edges[lower]++;
    rebuild(lower);
  }
  return e;
}

void dyn_con::del(edge e)
// delete the edge e
{
#ifdef STATISTICS
  n_del++;
#endif

  // if e is not an edge in F
  if(!tree_edge(e)) delete_non_tree(e);
  else
  // e is a tree edge
  {
    // remember e
    int e_level = level(e);
    node u = source(e);
    node v = target(e);

    // remove e
    delete_tree(e);

    // delete specific information for tree edges stored at e
    for(int j=0; j<=max_level; j++) delete[] (*Gp)[e]->tree_occ[j];
    delete[] (*Gp)[e]->tree_occ;

    // look for a replacement edge
    replace(u,v,e_level);
  }

  // delete information stored at e
  delete (*Gp)[e];
  (*Gp)[e] = nil;
  Gp->del_edge(e);
}

bool dyn_con::connected(node u, node v)
// return true if u and v are connected in the current graph
// and false otherwise
{
#ifdef STATISTICS
  n_query++;
#endif

  return connected(u,v,max_level);
}

void dyn_con::print_statistics(ostream& out)
{
#ifdef STATISTICS
  out << "\nStatistics\n==========\n";
  out << "number of nodes: " << Gp->number_of_nodes();
  out << " final number of edges: " << Gp->number_of_edges() << "\n\n";
  out << "user supplied operations\n";
  out << "number of ins operations: " << n_ins << "\n";
  out << "number of del operations: " << n_del << "\n";
  out << "number of connected operations: " << n_query << "\n\n";
  out << "internal variables\n";
  out << "number of levels: " << max_level+1 << "\n";
  out << "bound for rebuilds on highest level: " << rebuild_bound[max_level];
  out << "\nsmall_weight: " << small_weight << "\n";
  out << "maximum number of edges to sample: " << edges_to_sample << "\n";
  out << "small_set: " << small_set << "\n\n";
  out << "internal functions\n";
  out << "number of connected operations: " << n_connected << "\n";
  out << "number of insert_tree operations: " << n_ins_tree << "\n";
  out << "number of delete_tree operations: " << n_del_tree << "\n";
  out << "number of replace operations: " << n_replace << "\n";
  out << "  weight of T_1 too small: " << rep_small_weight << "\n";
  out << "  case 2.(b): " << rep_succ << "\n";
  out << "  case 3.(b): " << rep_big_cut << "\n";
  out << "  case 3.(c): " << rep_sparse_cut << "\n";
  out << "  case 3.(d): " << rep_empty_cut << "\n";
  out << "number of sample_and_test operations: " << n_sample_and_test << "\n";
  out << "number of get_cut_edges operations without recursive calls: ";
  out << n_get_cut_edges << "\n";
  out << "number of insert_non_tree operations: " << n_ins_non_tree << "\n";
  out << "number of delete_non_tree operations: " << n_del_non_tree << "\n";
  out << "number of move_edges: " << n_move_edges << "\n";
  out << "number of edges moved up: " << edges_moved_up << "\n";
  out << "number of edges moved down: " << edges_moved_down << "\n";
#else
  out << "\ndyn_con::print_statistics: sorry, no statistics available\n";
  out << "  compile libdc.a with -DSTATISTICS to get statistics\n\n";
#endif
}



bool dyn_con::connected(node x, node y, int i)
// Return true if x and y are connected on level i. Otherwise
// return false.
{
#ifdef STATISTICS
  n_connected++;
#endif

  // get the active occurrences of x and y at level i
  et_node x_act_occ = (*Gp)[x]->act_occ[i];
  et_node y_act_occ = (*Gp)[y]->act_occ[i];

  // return whether they belong to the same tree at level i
  return (x_act_occ->find_root() == y_act_occ->find_root());
}

void dyn_con::insert_tree(edge e, int i, bool create_tree_occs)
// Insert e into F_i. If create_tree_occs is true, then the storage for the
// tree_occ array for e is allocated.
{
#ifdef STATISTICS
  n_ins_tree++;
#endif

  // find the endpoints of e
  node u = source(e);
  node v = target(e);

#ifdef DEBUG
  cout << "(" << index(u) << "," << index(v) << ") tree ins at level ";
  cout << i << "\n";
#endif

  // enter level of e
  (*Gp)[e]->level = i;

  // create tree_occ array for e if requested
  if(create_tree_occs)
  {
    (*Gp)[e]->tree_occ = new et_node*[max_level+1];
    for(int lev=0; lev<=max_level; lev++)
    {
      (*Gp)[e]->tree_occ[lev] = new et_node[4];
      for(int j=0; j<4; j++) (*Gp)[e]->tree_occ[lev][j] = nil;
    }
  }

  // link the et_trees containing the active occurrences of u and v
  for(int j=i; j<=max_level; j++) et_link(u,v,e,j,this);

  // append e to the list of tree edges at level i
  (*Gp)[e]->tree_item = tree_edges[i].append(e);
}

void dyn_con::delete_tree(edge e)
// Remove the tree edge e from F.
{
#ifdef STATISTICS
  n_del_tree++;
#endif

  // get the level of e
  int i = level(e);

#ifdef DEBUG
  cout << "(" << index(source(e)) << "," << index(target(e)) << ") tree del ";
  cout << "at level " << i << "\n";
#endif

  // cut the spanning trees
  for(int j=i; j<=max_level; j++) et_cut(e,j,this);

  // remove e from the list of tree edges at level i
  tree_edges[i].del_item((*Gp)[e]->tree_item);

  // set tree_item of e to nil
  (*Gp)[e]->tree_item = nil;
}

void dyn_con::replace(node u, node v, int i)
// try to reconnect the trees on level i containing u and v.
// if not possible try to recurse on higher level
{
#ifdef STATISTICS
  n_replace++;
#endif

  // determine the level i trees containing u and v
  et_tree t1 = (*Gp)[u]->act_occ[i]->find_root();
  et_tree t2 = (*Gp)[v]->act_occ[i]->find_root();

  // let t1 be the smaller tree
  if(t1->get_subtree_weight() > t2->get_subtree_weight()) t1 = t2;

  int handle_cut = false;
  if(t1->get_subtree_weight() > small_weight)
  {
    // sample randomly at most edges_to_sample edges
    int not_done = true;
    edge e;
    for(int j=0; not_done && (j<edges_to_sample); j++)
    {
      e = sample_and_test(t1,i);
      if(e) not_done = false;
    }

    if(e)
    {
      // sampling was successful, insert e as a tree edge at level i
      delete_non_tree(e);
      insert_tree(e,i,true);
#ifdef STATISTICS
      rep_succ++;
#endif
    }
    else // sampling not successful
    {
      handle_cut = true;
    }

  }
  else // weight of T_1 too small (too few adjacent edges)
  {
    handle_cut = true;
#ifdef STATISTICS
    rep_small_weight++;
#endif
  }

  if(handle_cut)
  // sampling was unsuccessful or too few edges
  {
    // determine all edges crossing the cut
    list<edge> cut_edges;
    if(t1->get_subtree_weight() > 0)
    {
      get_cut_edges(t1,i,cut_edges);
#ifdef STATISTICS
      n_get_cut_edges++;
#endif
    }

    if(cut_edges.size() == 0)
    // no replacement edge on this level, recurse on higher level if possible
    {
#ifdef STATISTICS
      rep_empty_cut++;
#endif
      if(i<max_level) replace(u,v,i+1);
    }
    else  // cut_edges.size() > 0
    {
      if(cut_edges.size() >= t1->get_subtree_weight()/small_set)
      // if cut_edges is big enough we reconnect t1 and t2 on level i
      {
#ifdef STATISTICS
        rep_big_cut++;
#endif
        edge reconnect = cut_edges.pop();
        delete_non_tree(reconnect);
        insert_tree(reconnect,i,true);
      }
      else
      {
        // 0 < cut_edges.size() < t1->get_subtree_weight()/small_set
        // there are too few edges crossing the cut
#ifdef STATISTICS
        rep_sparse_cut++;
#endif
        edge reconnect = cut_edges.pop();
        delete_non_tree(reconnect);

        if(i<max_level)
        {
          // move cut edges one level up
          insert_tree(reconnect,i+1,true);
          added_edges[i+1]++;
          edge e;
          forall(e,cut_edges)
          {
            delete_non_tree(e);
            insert_non_tree(e,i+1);
            added_edges[i+1]++;
          }
#ifdef STATISTICS
          edges_moved_up += cut_edges.size() + 1;
#endif
          rebuild(i+1);
        }
        else     // on top level, no moving of edges
        {
          insert_tree(reconnect,i,true);
        }
      }
    }
  }
}

edge dyn_con::sample_and_test(et_tree T, int i)
// Randomly select a non-tree edge of G_i that has at least one
// endpoint in T, where an edge with both endpoints in T is picked
// with 2/w(T) and an edge with exactly one endpoint in T is picked
// with probability 1/w(T).
// Test if exactly one endpoint is in T, and if so return the edge.
// Otherwise return nil.
{
#ifdef STATISTICS
  n_sample_and_test++;
#endif

  // get the number of adjacencies
  int no_of_adj = T->get_subtree_weight();

  // pick a random one
  int rnd_adj = rs(1,no_of_adj);

  // locate the et_node representing this adjacency and get the corr. node
  int offset;
  et_node et_repr = et_locate(T,rnd_adj,offset);
  node u = et_repr->get_corr_node();

  // locate the edge corresp. to offset adjacent to u at level i
  ed_node en = ed_locate((*Gp)[u]->adj_edges[i],offset,offset);
  edge e = en->get_corr_edge();

  // get the second node of e
  node v = (source(e) == u) ? target(e) : source(e);

  // if v is in a different tree at level i then return e else nil
  if(connected(u,v,i)) return nil;
  else                 return e;
}

void dyn_con::traverse_edges(ed_node ed, list<edge>& edge_list)
// append edges with exactly one endpoint in subtree rooted at ed to edge_list
// auxiliary function called by get_cut_edges
{
  if(ed)
  {
    edge e = ed->get_corr_edge();
    if(!connected(source(e),target(e),level(e)))
    {
      // only one endpoint of e in current spanning tree -> append edge
      edge_list.append(e);
    }

    traverse_edges(ed->left_child(),edge_list);
    traverse_edges(ed->right_child(),edge_list);
  }
}

void dyn_con::get_cut_edges(et_node u, int level, list<edge>& result)
// Return the edges with exactly one endpoint in the et_tree rooted at u
// in result.
{
  if(u && u->get_subtree_weight())
  {
    node v = u->get_corr_node();
    if(u->is_active()) traverse_edges((*Gp)[v]->adj_edges[level],result);
    get_cut_edges(u->left_child(),level,result);
    get_cut_edges(u->right_child(),level,result);
  }
}

void dyn_con::insert_non_tree(edge e, int i)
// Insert the non-tree edge e into G_i.
{
#ifdef STATISTICS
  n_ins_non_tree++;
#endif

#ifdef DEBUG
  cout << "(" << index(source(e)) << "," << index(target(e));
  cout << ") non-tree ins at level " << i << "\n";
#endif

  (*Gp)[e]->level = i;
  node u = source(e);
  node v = target(e);

  // insert e in the adjacency trees of its endpoints at level i
  (*Gp)[e]->non_tree_occ[0] =
                   ed_insert((*Gp)[u]->adj_edges[i],e,ed_dummy);
  (*Gp)[e]->non_tree_occ[1] =
                   ed_insert((*Gp)[v]->adj_edges[i],e,ed_dummy);

  // update non_tree_edges[i]
  (*Gp)[e]->non_tree_item = non_tree_edges[i].append(e);

  // increase the weight of the active occurrences of u and v at level i
  (*Gp)[u]->act_occ[i]->add_weight(1);
  (*Gp)[v]->act_occ[i]->add_weight(1);
}

void dyn_con::delete_non_tree(edge e)
// Delete the non-tree edge e.
{
#ifdef STATISTICS
  n_del_non_tree++;
#endif

  // find the endpoints and the level of e
  node u = source(e);
  node v = target(e);
  int i = level(e);

#ifdef DEBUG
  cout << "(" << index(source(e)) << "," << index(target(e));
  cout << ") non-tree del at level " << i << "\n";
#endif

  // remove e from the ed_trees of u and v at level i
  ed_delete((*Gp)[u]->adj_edges[i],(*Gp)[e]->non_tree_occ[0],ed_dummy);
  (*Gp)[e]->non_tree_occ[0] = nil;
  ed_delete((*Gp)[v]->adj_edges[i],(*Gp)[e]->non_tree_occ[1],ed_dummy);
  (*Gp)[e]->non_tree_occ[1] = nil;

  // remove e from the list of non-tree edges at level i
  non_tree_edges[i].del_item((*Gp)[e]->non_tree_item);
  (*Gp)[e]->non_tree_item = nil;

  // decrease the weights of the active occurrences of u and v if they exist
  (*Gp)[u]->act_occ[i]->add_weight(-1);
  (*Gp)[v]->act_occ[i]->add_weight(-1);
}

void dyn_con::rebuild(int i)
// does a rebuild at level i if necessary
{
  // rebuilds take place only at level 3 and higher
  if(i<3) return;

  // count added edges at level j>=i
  int sum_added_edges = 0;
  for(int j=i; j<=max_level; j++) sum_added_edges += added_edges[j];

  if(sum_added_edges > rebuild_bound[i])
  {
#ifdef DEBUG
    cout << "rebuild(" << i << ")\n";
#endif
    // move edges down
    move_edges(i);
    for(int j=i; j<=max_level; j++) added_edges[j] = 0;
  }
}

void dyn_con::move_edges(int i)
// For j>=i, insert all edges of F_j into F_{i-1}, and all
// non-tree edges of G_j into G_{i-1}.
{
#ifdef STATISTICS
  n_move_edges++;
#endif

  // for each level starting at max_level and ending at i...
  for(int j=max_level; j>=i; j--)
  {
#ifdef STATISTICS
    edges_moved_down += non_tree_edges[j].size() + tree_edges[j].size();
#endif
    // move non-tree edges
    while(non_tree_edges[j].size())
    {
      edge e = non_tree_edges[j].head();
      // delete non-tree edge at level j ...
      delete_non_tree(e);
      // ... and insert it into level i-1
      insert_non_tree(e,i-1);
    }

    // move tree edges
    while(tree_edges[j].size())
    {
      edge e = tree_edges[j].head();

      // update tree_edges[j], tree_edges[i-1], tree_item and level
      tree_edges[j].del_item(Gp->inf(e)->tree_item);
      Gp->inf(e)->tree_item = tree_edges[i-1].append(e);
      Gp->inf(e)->level = i-1;

      // link the corresponding et_trees from level i-1 to j-1
      for(int k=i-1; k<j; k++)
      {
        et_link(source(e),target(e),e,k,this);
      }
    }
  }
}

dyn_con::dyn_con(dc_graph& G, int ml_reb_bound, int n_levels,
                 int edges_to_samp, int small_w, int small_s)
// constructor, initializes the dynamic connectivity data structure
// if ml_reb_bound >= 1 it specifies rebuild_bound[max_level] (default is 5000)
// if n_levels > 0 then it specifies the number of levels (default O(log n))
// if edges_to_samp >= 0 then it specifies edges_to_sample (default 16 log^2 n)
// if small_w >= 0 then it specifies small_weight (default log^2 n)
// if small_s >= 1 then it specifies small_set (default 8 log n)
{
  // --- initialize the constants ---
  Gp = &G;
  int log_n = 0;
  int i;
  for(i = G.number_of_nodes(); i; i /= 2) log_n++;

  if(small_w>=0) small_weight = small_w;
  else           small_weight = log_n * log_n;

  if(edges_to_samp>=0) edges_to_sample = edges_to_samp;
  else                 edges_to_sample = 16 * log_n * log_n;

  if(small_s>=1) small_set = small_s;
  else           small_set = 8 * log_n;

  if(n_levels > 0) max_level = n_levels - 1;
  else
  {
    max_level = 2 * log_n;
    for(int k=4; (k<ml_reb_bound) && (max_level>=2); k *= 2, max_level--);
  }

#ifdef DEBUG
  cout << "|V(G)|          = " << G.number_of_nodes() << "\n";
  cout << "max_level       = " << max_level << "\n";
  cout << "edges_to_sample = " << edges_to_sample << "\n";
  cout << "small_set       = " << small_set << "\n\n";
#endif

  // --- initialize dummy nodes ---
  et_dummy = new et_node_struct(this,nil);
  ed_dummy = new ed_node_struct(nil);

  // --- initialize the edge lists ---
  non_tree_edges = new list<edge> [max_level+1];
  tree_edges = new list<edge> [max_level+1];

  // --- initialize added_edges ---
  added_edges = new int[max_level+1];
  for(i=0; i<=max_level; i++) added_edges[i] = 0;

  // --- initialize rebuild_bound ---
  rebuild_bound = new int[max_level+1];
  int bound;
  if(ml_reb_bound>=1) bound = ml_reb_bound;
  else                bound = 5000;
  for(int k=max_level; k>=0; k--)
  {
    rebuild_bound[k] = bound;
    if(bound < MAXINT/2) bound *= 2;  // double the bound if possible
  }

  // --- initialize the nodes ---
  node u;
  forall_nodes(u,G)
  {
    G[u] = new dc_node_struct();
    G[u]->act_occ = new et_node[max_level+1];
    G[u]->adj_edges = new ed_tree[max_level+1];
    for(i=0; i<=max_level; i++)
    {
      G[u]->act_occ[i] = new et_node_struct(this,u,i,true);
      G[u]->adj_edges[i] = nil;
    }
  }

#ifdef STATISTICS
  // --- initialize variables touched by initializing edges ---
  n_connected = 0;
  n_ins_tree = 0;
  n_ins_non_tree = 0;
#endif

  // --- initialize the edges ---
  edge e;
  forall_edges(e,G)
  {
    G[e] = new dc_edge_struct();
    if(!connected(source(e),target(e),0)) insert_tree(e,0,true);
    else                                  insert_non_tree(e,0);
  }

#ifdef STATISTICS
  // --- initialize statistics ---
  n_ins = 0;
  n_del = 0;
  n_query = 0;
  n_connected = 0;
  n_ins_tree = 0;
  n_del_tree = 0;
  n_replace = 0;
  rep_succ = 0;
  rep_big_cut = 0;
  rep_sparse_cut = 0;
  rep_empty_cut = 0;
  rep_small_weight = 0;
  n_sample_and_test = 0;
  n_get_cut_edges = 0;
  n_ins_non_tree = 0;
  n_del_non_tree = 0;
  n_move_edges = 0;
  edges_moved_up = 0;
  edges_moved_down = 0;
#endif
}

dyn_con::~dyn_con()
{
  // first delete all edges in the data structure (not in G)
  edge e;
  forall_edges(e,*Gp)
  {
    if(tree_edge(e))
    {
      delete_tree(e);
      for(int j=0; j<=max_level; j++) delete[] (*Gp)[e]->tree_occ[j];
      delete[] (*Gp)[e]->tree_occ;
    }
    else delete_non_tree(e);

    delete (*Gp)[e];
    (*Gp)[e] = nil;
  }

  // delete fields (edge lists are empty, no need to clear() them)
  delete[] non_tree_edges;
  delete[] tree_edges;
  delete[] added_edges;
  delete[] rebuild_bound;

  // delete the et_nodes and the information at the nodes of G
  node v;
  forall_nodes(v,*Gp)
  {
    // per node of G only its active occurrence at each level is left
    for(int i=0; i<=max_level; i++) delete (*Gp)[v]->act_occ[i];

    delete[] (*Gp)[v]->act_occ;
    delete[] (*Gp)[v]->adj_edges;

    delete (*Gp)[v];
    (*Gp)[v] = nil;
  }

  // delete dummy nodes
  delete et_dummy;
  delete ed_dummy;
}
