// ------------------------------------------------------------------- //
// check_spars.c:                                                      //
// check_spars is a small test program for the sparsification dynamic  //
// connectivity data structure.                                        //
//                                                                     //
// usage:                                                              //
//          check_spars n ti                                           //
// where n is the size of a test graph, and ti the test interval.      //
//                                                                     //
// check_spars generates a graph with n nodes, the original graph OG.  //
// It maintains a second graph G which is always (isomorphic to) a     //
// subgraph of OG. Initially, G is isomorphic to OG. Then random edges //
// in OG are selected. If their counterparts in G exist they are       //
// deleted. If their counterparts do not exist they are inserted.      //
// After every ti updates all possible queries are performed and       //
// compared to the answers computed by a static algorithm. After every //
// such test a dot is printed.                                         //
//                                                                     //
// Quit the program with CTRL-C.                                       //
// ------------------------------------------------------------------- //

#include<STree.h>
#include<extalg.h>
#include<ed_tree.h>
#include<LEDA/graph_alg.h>

void ohno()
{ cout << "OHH, NOOOOO!!!!!!\n"; }


bool HeuristicOn = true;

main(int argc, char* argv[])
{
  if(argc != 3)
  {
    cout << "See check_spars.c for some documentation.\n";
    return 5;
  }

  int size = atoi(argv[1]);            // size of original graph
  int test_interval = atoi(argv[2]);   // test the d.s. after every so many upd.

  graph OG;                            // original graph
  graph G;                             // current subgraph
  ed_tree OG_edges = nil;              // edges in original graph,
                                       // allows fast random sampling
  triangulated_planar_graph(OG,size);  // or complete_graph, grid_graph, ...

  // the following two arrays represent the graph isomorphism between OG and G
  edge_array<edge> corr_edge(OG,nil);  // corresponding edge in current subgraph
  node_array<node> corr_node(OG,nil);  // corresponding node

  // clone OG
  node v;
  forall_nodes(v,OG) corr_node[v] = G.new_node();
  ed_node dummy = new ed_node_struct(nil);
  edge e;
  forall_edges(e,OG)
  {
    ed_insert(OG_edges,e,dummy);
    corr_edge[e] = G.new_edge(corr_node[source(e)],corr_node[target(e)]);
  }

  int n = OG.number_of_nodes();
  Sparse dc(&G, SpanningTree, n, n-1, HeuristicOn, false );

  int m = OG.number_of_edges();
  random_source S(1,m);

  int rnd_ind, offset;
  int number_upd = 0;
  while(true)
  {
    // Pick a random edge in OG and insert it in G if it is not already there.
    // If it is already there delete it in G.
    S >> rnd_ind;
    ed_node ed = ed_locate(OG_edges,rnd_ind,offset);
    e = ed->get_corr_edge();
    if(corr_edge[e])
    {
      G.del_edge(corr_edge[e]);
      dc.DeleteEdge(corr_edge[e],1);
      corr_edge[e] = nil;
    }
    else
    {
      corr_edge[e] = G.new_edge(corr_node[source(e)],corr_node[target(e)]);
      dc.InsertEdge(corr_edge[e],1);
    }
    number_upd++;

    // Check whether the d.s. answers all queries as it should answer them.
    if(number_upd == test_interval)
    {
      node_array<int> comp_num(G);
      COMPONENTS(G,comp_num);

      node u;
      forall_nodes(u,G)
      {
        node v;
        forall_nodes(v,G)
        {
          int connected = (comp_num[u] == comp_num[v]);
          if(connected != dc.connected(u,v))
          {
	    G.print_node(u); G.print_node(v); newline;
            ohno();
          }
        }
      }
      cout << "." << flush;
      number_upd = 0;
    }

  }
}


