// ---------------------------------------------------- //
// gen_ops.c: We generate an initial random subgraph of //
//            a given graph and a random sequence of    //
//            operations and write them to files.       //
//                                                      //
// format of operations file:                           //
//            <#operations>                             //
//            0 33 10                 <-- insertion     //
//            1 12 42                 <-- deletion      //
//            2 3 17                  <-- query         //
//                                                      //
// usage: gen_ops graph m name #ins #del #queries       //
//                                                      //
//        where graph is a file containing the graph of //
//        all allowed edges, m is the number of edegs   //
//        in the initial random subgraph, #ins, #del,   //
//        #queries specify the numbers of insertions,   //
//        deletions, and queries, resp.                 //
//                                                      //
// The initial graph is written to name.gr and the      //
// operations are written to name.upd.                  //
//                                                      //
// Written by David Alberts				//
// ---------------------------------------------------- //
#include<ed_tree.h>
#include<sys/time.h>
#include<LEDA/graph.h>
#include<LEDA/graph_alg.h>
#include<fstream.h>

graph TG;				// the test graph

// we implicitly maintain a subgraph by viewing some edges
// as edges of the current subgraph and the rest as non-edges
// they are stored in balanced binary trees for fast random access
ed_tree SG_edges = nil;			// edges in current subgraph
ed_tree SG_non_edges = nil; 		// edges in TG \ SG
ed_node dummy;

// block size
unsigned int block_size = 500;

random_source RS;

edge random_ins()
// make a random non-edge in the current graph an edge and return it
{
  int rnd_index = RS(1,SG_non_edges->get_subtree_weight());
  int offset;
  ed_node ed = ed_locate(SG_non_edges,rnd_index,offset);
  edge e = ed->get_corr_edge();
  ed_delete(SG_non_edges,ed,dummy);
  ed_insert(SG_edges,e,dummy);
  return e;
}

edge random_del()
// make a random edge in the current graph a non-edge and return it
{
  int rnd_index = RS(1,SG_edges->get_subtree_weight());
  int offset;
  ed_node ed = ed_locate(SG_edges,rnd_index,offset);
  edge e = ed->get_corr_edge();
  ed_delete(SG_edges,ed,dummy);
  ed_insert(SG_non_edges,e,dummy);
  return e;
}

// the type of an operation
enum dc_ops_type { ins=0, del=1, query=2 };

main(int argc, char* argv[])
{
  if(argc < 7 || argc > 8)
  {
    cerr << "gen_ops: usage is\n\tgen_ops graph m name #ins #del #queries [seed]\n";
    return 5;
  }

  // ----- read in the graph and the number of operations -----
  TG.read(argv[1]);
  int n = TG.number_of_nodes();
  int m = atoi(argv[2]);
  int n_ins = atoi(argv[4]);
  int n_del = atoi(argv[5]);
  int n_que = atoi(argv[6]);
  int n_ops = n_ins + n_del + n_que;

  // ----- store the edges in a balanced binary tree for -----
  // ----- random selection                              -----
  dummy = new ed_node_struct(nil);
  edge e;
  forall_edges(e,TG) ed_insert(SG_non_edges,e,dummy);

  // ----- initialize random number geneator
  if( argc == 8 )   RS.set_seed( atoi(argv[7]) );

  // ----- create initial random subgraph -----
  // ----- and write it to disk -----
  // we have to copy it, because it is only implicit
  graph dummy_gr;
  node* nodes = new node[n];
  int i;
  for(i=0; i<n; i++) nodes[i] = dummy_gr.new_node();
  for(i=0; i<m; i++)
  {
    e = random_ins();     // this creates subgraph implicitly
    dummy_gr.new_edge(nodes[index(source(e))],nodes[index(target(e))]);
  }
  dummy_gr.write(string("%s.gr",argv[3]));
  dummy_gr.clear();


  // ----- compute a random sequence of operations -----
  ofstream out(string("%s.upd",argv[3]));
  out << n_ops << "\n";        // the file starts with #ops

  while(n_ins+n_del+n_que)
  {
    int rnd = RS(0,n_ins+n_del+n_que-1);
    if(rnd < n_ins)
    {
      if(SG_non_edges)
      {
	edge e = random_ins();
	out << ins << " " << index(source(e)) << " ";
	out << index(target(e)) << "\n";
	n_ins--;
      }
      else // no non-edges left
      {
	if(n_del) // try a deletion
	{
	  edge e = random_del();
	  out << del << " " << index(source(e)) << " ";
	  out << index(target(e)) << "\n";
	  n_del--;
	}
	else // impossible choice of insertions and deletions
	{
	    out << "error error error error\n";
	    cerr << "Sorry, Dave. I'm afraid I can't do that.\n";
	}
      }
    }
    else  // rnd >= n_ins
    {
      if(rnd < n_ins + n_del)
      {
	if(SG_edges)
	{
	  edge e = random_del();
	  out << del << " " << index(source(e)) << " ";
	  out << index(target(e)) << "\n";
	  n_del--;
	}
	else
	{
	  if(n_ins)
	  {
	    edge e = random_ins();
	    out << ins << " " << index(source(e)) << " ";
	    out << index(target(e)) << "\n";
	    n_ins--;
	  }
	  else // impossible choice of insertions and deletions
	  {
	    out << "error error error error\n";
	    cerr << "Sorry, Dave. I'm afraid I can't do that.\n";
	  }
	}
      }
      else // rnd >= n_ins + n_del
      {
	out << query << " " << RS(0,n-1) << " ";
	out << RS(0,n-1) << "\n";
	n_que--;
      }
    }
  }

  // clean up
  delete dummy;
  delete[] nodes;
}

