/******************************************************************************

 Manage hash list of "symbols" 

 Code adapted from William Cohen by Yoram Singer

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* for strcmp, strcpy */
#include "hash.h"

#define newmem(n,type)  ((type *) safe_calloc((n),sizeof(type)))

/* hash table sizes... */
static int Sizes[] = {5003,100003,2000003,40000003,0};
static int S = 0;

/* global data managed in this file */
typedef struct hash_tab_entry_s {
    Hash_Entry data;
    struct hash_tab_entry_s *next;
} hash_tab_entry_t; 

static hash_tab_entry_t **Hash_tab=NULL;

static int hash(char *);
static hash_tab_entry_t *findslot(char *,int);
static hash_tab_entry_t *addslot(char *,int);



/*VARARGS
static void fatal(fmt,a1,a2,a3,a4,a5,a6)
char *fmt;
{
    fprintf(stderr,"fatal: ");
    fprintf(stderr,fmt,a1,a2,a3,a4,a5,a6);
    fprintf(stderr,"\n");
    exit(1);
}
*/

static char *safe_calloc(int n, int size)
     //int n,size;
{
    char *space;

    space = (char *)calloc((unsigned)n,(unsigned)size);
    if (space==NULL) {
      printf("out of memory"); exit(1);
    }
    return space;
}



/*VARARGS
static void warning(fmt,a1,a2,a3,a4,a5,a6)
char *fmt;
{
    fprintf(stderr,"warning: ");
    fprintf(stderr,fmt,a1,a2,a3,a4,a5,a6);
    fprintf(stderr,"\n");
}
*/


static void show_symbol(Hash_Entry *s) /* debugging print function */
     //Hash_Entry *s;
{
    printf("<count=%4d symbol '%s'>",(int) s->info, s->name);
}



/*****************************************************************************
 *
 *	map a string to the corresponding symbol, creating one if needed
 *
 *****************************************************************************/
Hash_Entry *Hash_Insert(char *str, Hash_Info *hi) 
{
    int n_hashed;
    hash_tab_entry_t *slot,*nextslot;
    hash_tab_entry_t **old_hash_tab;
    int index;
    int i;

	n_hashed = hi->n_hashed;
	S = hi->S;
	Hash_tab = (hash_tab_entry_t **) hi->Hash_tab;

    /* initialize hash table if necessary */
    if (Hash_tab==NULL) {
	Hash_tab = newmem(Sizes[S],hash_tab_entry_t *);
    }

    index = hash(str);
    if ((slot=findslot(str,index))==NULL) {
	n_hashed++;
	slot = addslot(str,index);
    } 

    /* if not found, add to table */
    if (n_hashed == 2*Sizes[S]) {
	if (Sizes[S+1]==0) {
	    printf("hash table is crowded, performance may suffer");
	} else {
	    /* warning("hash table is too crowded...rehashing"); */
	    old_hash_tab = Hash_tab;
	    Hash_tab = newmem(Sizes[++S],hash_tab_entry_t *);
	    for (i=0; i<Sizes[S-1]; i++) {
		for (slot=old_hash_tab[i]; slot!=NULL; slot=nextslot) {
		    nextslot = slot->next;
		    index = hash(slot->data.name);
		    slot->next = Hash_tab[index];
		    Hash_tab[index] = slot;
		}
	    }
	    /* warning("finished rehashing"); */
	    slot=findslot(str,hash(str));
	}
    }

	hi->n_hashed = n_hashed;
	hi->S = S;
	hi->Hash_tab = (void *) Hash_tab;

    return &(slot->data);
}



/*****************************************************************************
 *
 *	map a string to the corresponding symbol, creating one if needed
 *
 *****************************************************************************/
Hash_Entry *Hash_Find(char *str, Hash_Info *hi) 
{
    int n_hashed;
    hash_tab_entry_t *slot;
    int index;

	n_hashed = hi->n_hashed;
	S = hi->S;
	Hash_tab = (hash_tab_entry_t **) hi->Hash_tab;

	if (Hash_tab == NULL)
		return NULL;

    index = hash(str);
    if ((slot=findslot(str,index))!=NULL)
		return &(slot->data);
	else
		return NULL;
}



static int hash(char *cp)
{
    unsigned int index=0;

    /* compute hash function */
    while (*cp) index = ((index << 7) + (index << 3) + index) ^ *cp++;
    return (index%Sizes[S]);
}



static hash_tab_entry_t *findslot(char *str,int index)
{
    hash_tab_entry_t *slot;

    /* look up in table */
    for (slot=Hash_tab[index]; slot!=NULL; slot=slot->next) {
	if (strcmp(slot->data.name,str)==0) {
	    return slot;
	}
    }
    return NULL;
}



static hash_tab_entry_t *addslot(char *str,int index)
{
    hash_tab_entry_t *slot;

    /* if not found, add to table */
    slot = newmem(1,hash_tab_entry_t);
    slot->next = Hash_tab[index];
    Hash_tab[index] = slot;
    slot->data.name = newmem(strlen(str)+1, char);
    strcpy(slot->data.name,str);
    slot->data.info = NULL;

    return slot;
}



void show_hashtab(Hash_Info *hi)
{
    int i;
    hash_tab_entry_t *slot;

	S = hi->S;
	Hash_tab = (hash_tab_entry_t **) hi->Hash_tab;

    for (i=0; i<Sizes[S]; i++) {
	if (Hash_tab[i]) {
	    printf("slot %7d: ",i);
	    for (slot=Hash_tab[i]; slot!=NULL; slot=slot->next) {
		printf("   "); show_symbol(&(slot->data)); printf("\n");
	    }
	}
    }
}


void scan_hashtab(Hash_Info *hi)
{
	Hash_Entry *he;


	Init_Hash_Scan(hi);

	while ((he = Next_Hash_Entry()) != NULL) {
		show_symbol(he);
		printf("\n");
	}
}


static int cur_index;

void Init_Hash_Scan(Hash_Info *hi)
{
	cur_index = -1;
	S = hi->S;
	Hash_tab = (hash_tab_entry_t **) hi->Hash_tab;
}


Hash_Entry *Next_Hash_Entry()
{
    static hash_tab_entry_t *slot = NULL;
	hash_tab_entry_t *t_slot;


	if (slot !=NULL) {
		t_slot = slot;
		slot = slot->next;
		return &(t_slot->data);
	}

	cur_index++;

	while (cur_index < Sizes[S] && Hash_tab[cur_index] == NULL)
		cur_index++;

	if (cur_index >= Sizes[S])
		return NULL;

	t_slot = Hash_tab[cur_index];
	slot = t_slot->next;

	return &(t_slot->data);
}


/*
void Apply_To_Hash(void (*fn)(), Hash_Info *hi)
{

	Hash_Entry *he;

	Init_Hash_Scan(hi);

	while ((he = Next_Hash_Entry()) != NULL)
		(*fn)(he->name, he->info);
}



void Apply_To_Hash_Direct(void (*fn)(), Hash_Info *hi)
{

	Hash_Entry *he;

	Init_Hash_Scan(hi);

	while ((he = Next_Hash_Entry()) != NULL)
		(*fn)(he);
}

*/

Hash_Info *Init_Hash()
{
	Hash_Info *hi;

	hi = newmem(1,Hash_Info);

	hi->S = 0;
	hi->n_hashed = 0;
	hi->Hash_tab = NULL;

	return hi;
}

