// ---------------------------------------------------- //
// test_hdt.c: test the performance of the dynamic      //
//            connectivity data structure by HDT        //
//                                                      //
// usage: test_hdt name n_upd1 n_upd2 ...               //
//        after gen_ops was executed and produced       //
//        name.gr and name.upd.                         //
//	  Output performance after n_upd1 updates,	//
//	  after n_upd2 updates, etc.			//
//                                                      //
// test_hdt name reads name.gr as initial graph and     //
// performs the operations specified in name.upd.       //
// Output is average seconds per operations, averaged   //
// over the last performance sample.			//
//                                                      //
// output format:                                       //
//  n m pre_ti n_upd1 ti_upd1 sec/op1 ...		//
//	n = #nodes					//
//	m = #edges					//
//	pre_ti = preprocessing time (sec)		//
//	n_upd1 = (from input)				//
//	ti_upd1 = time to perform n_upd1 updates (sec)  //
//	sec/op1 = average seconds/operation, starting	//
//	          from previous n_upd (0 if n_upd1)	//
//	(These last three repeat for every n_upd input) //
//                                                      //
// Written by David Alberts, adapted by H. S. Rahul     //
// ---------------------------------------------------- //

#include <dc_hdt.h>
#include <fstream.h>
#include <hash.h>
#include <sys/times.h>
#include <limits.h>
#include <stdio.h> // to check if a file exists

char key_str[100];

Hash_Entry * insert_edge(edge e, int u, int v, Hash_Info *hi) {
  if (u > v) {
    int tmp = u;
    u = v;
    v = tmp;
  }
  // u < v here
  sprintf(key_str,"%d,%d", u, v);
  Hash_Entry *he = Hash_Insert(key_str, hi);
  if (he == NULL) return he;
  he->info = e;
  return he;
}

int find_edge(edge *e, int u, int v, Hash_Info *hi) {
  if (u > v) {
    int tmp = u;
    u = v;
    v = tmp;
  }
  // u < v here
  sprintf(key_str,"%d,%d", u, v);
  Hash_Entry *he = Hash_Find(key_str, hi);
  if (he == NULL) return 0;
  *e = he->info;
  return 1;
}

float current_time() 
{
  struct tms buff;
  times(&buff);
  return((buff.tms_utime*1.0)/CLK_TCK);
}

float elapsed_time(float start)
{
  struct tms buff;
  times(&buff);
  return ((buff.tms_utime*1.0)/CLK_TCK - start);
}


graph dummy;                  // this is needed for reading in the
                              // initial graph as I am using a parameterized
                              // graph :-(

dc_graph TG;                  // the test graph

main(int argc, char* argv[])
{
  if(argc <= 4)
  {
    cerr << argv[0] << ": usage is\n\t" << argv[0] 
	 << " name samp_threshold_edges base_size n_upd1 n_upd2 ...\n";
    return 5;
  } 
  
  int samp_threshold_edges = strtol(argv[2], NULL, 10);
  int base_size = strtol(argv[3], NULL, 10);

  int num_total_upd = argc - 4;
  int *target_total_upd = new int[num_total_upd+1];
  float *pro_tf = new float[num_total_upd+1];

  int max_target_upd = -1;

  for (int k = 0; k < num_total_upd; k++) {
    target_total_upd[k] = strtol(argv[4+k], NULL, 10);
    if (max_target_upd < target_total_upd[k]) {
      max_target_upd = target_total_upd[k];
    } else {
      cout << "WARNING: Updates should only be given in ascending order\n";
    }
  }

  /* TEST CODE, don't know how to write it in C++ */
  string fname("%s.gr",argv[1] );
  FILE *F = fopen( fname, "r");
  if( ! F )
  {
      cout << "Can't open graph file '" << fname << "'\n";
      return 6;
  }      
  fclose( F );
  fname = string("%s.upd",argv[1] );
  F = fopen( fname, "r");
  if( ! F )
  {
      cout << "Can't open update file '" << fname << "'\n";
      return 6;
  }      
  fclose( F );
  /* END TEST */

  Hash_Info *hi = Init_Hash();

  // ----- initialize the test graph -----
  // nodes contains the nodes of TG by index 
  // edges is the adj. matrix of TG          - for performing updates

  dummy.read(string("%s.gr",argv[1]));

  int n = dummy.number_of_nodes();
  int m = dummy.number_of_edges();
  node* nodes = new node[n];
  node v;
  int i = 0;

  forall_nodes(v,dummy) nodes[i++] = TG.new_node();
  edge e;
  forall_edges(e,dummy)
  {
    insert_edge(TG.new_edge(nodes[index(source(e))],nodes[index(target(e))]), 
		index(source(e)), index(target(e)), hi);
  }
  dummy.clear();

  ifstream in(string("%s.upd",argv[1]));

  // ----- set up arrays containing update type and affected edge -----
  int n_upd;                        // the number of updates
  in >> n_upd;

  int* update = new int[n_upd+1];     // the type of update
  int* upd_source = new int[n_upd+1]; // the indices of the involved nodes
  int* upd_target = new int[n_upd+1];

  // ----- read in the update sequence -----
  i = 0;
  //cout << "About to read updates\n" << flush;
  while(!in.eof())
  {
    int op;
    in >> op;
    update[i] = op;
    int src, tg;
    in >> src >> tg;
    upd_source[i] = src;
    upd_target[i] = tg;
    i++;
  }
    
  // ----- initialize the data structure -----
  float pre_ti = current_time();	// prepr. time
  hdt_base *dc;
  dc = new hdt_base(TG, -1, samp_threshold_edges, base_size); 
  pre_ti = elapsed_time(pre_ti);
  
  // ----- perform the updates -----
  int conn_count = 0;         // #queries returning true
  float pro_ti = current_time();

#ifndef STATISTICS
  cout << n << " " << m << " " << string("%.2f ", pre_ti);
#endif

  int count_upd = 0;
  n_upd = (n_upd <= max_target_upd) ? n_upd : max_target_upd;

  for(int j = 0; j < n_upd; j++)  // i is one too big
  {
    if (j == target_total_upd[count_upd]) {
      pro_tf[count_upd] = elapsed_time(pro_ti);
      count_upd++;
      pro_ti = current_time();

    }

    switch(update[j])
    {
      case 0:      // insertion
      {
        insert_edge(dc->ins(nodes[upd_source[j]],nodes[upd_target[j]]),
		    upd_source[j], upd_target[j], hi);
        break;
      }
      case 1:      // deletion
      {
        if (find_edge(&e, upd_source[j], upd_target[j], hi)) {
	  dc->del(e);
	} else {
	  cout << "Error: Could not find edge (" << upd_source[j] <<
	    ", " << upd_target[j] << "\n";
	}
	break;
      }
      case 2:      // query
      {
        if(dc->connected(nodes[upd_source[j]],nodes[upd_target[j]]))
          conn_count++;
        break;
      }
    }
  }

  pro_tf[count_upd] = elapsed_time(pro_ti);
  target_total_upd[count_upd] = n_upd;
  count_upd++;

#ifndef STATISTICS
  i=0;
  cout << string("%d %.2f %.5f ",target_total_upd[i], pro_tf[i], 
		 (pro_tf[i]/(target_total_upd[i])));
  for (i = 1; i < count_upd; i++) {
    cout << string("%d %.2f %.5f ",target_total_upd[i], pro_tf[i], 
		   (pro_tf[i]/(target_total_upd[i]-target_total_upd[i-1])));
  }
  cout << "\n";
#endif
  
#ifdef STATISTICS
  dc->print_statistics(cout);
  cout << "\n" << flush;
#endif
}


