#ifndef _DC_HDT_H
#define _DC_HDT_H

// dc_hdt.h

// ------------------------------------------------------------- //
// class dc_hdt							 //
//								 //
// This is an implementation of the dynamic connectivity	 //
// algorithm by Holm, de Lichtenberg, and Thorup.		 //
//                                                               //
// The code is a modified version of David Albert's		 //
// implementation of the Henzinger-King algorithm.		 //
//                                                               //
// Version 1.0 (July, 2000)                                      //
//								 //
// Implementation by Raj D. Iyer and H. S. Rahul		 //
// There is ABSOLUTELY NO WARRANTY.                              //
// ------------------------------------------------------------- //

#include <hdt_rnbw_tree.h>
#include<LEDA/basic.h>
#include<LEDA/list.h>
#include<LEDA/graph.h>
// Global declarations

// ------------ et_trees ----------------- //
// some forward definitions
class hdt_base;

class   et_node_struct;
typedef et_node_struct* et_node;
typedef et_node et_tree;


class et_node_struct : public hdt_rnbw_node_struct {

public:
  // Constructors
  et_node_struct(hdt_base* dcp, node v, int my_level = -1, 
				 int activate = false);
  // Create a new et_node_struct at level my_level for v.
  // Activate it if activate is true.
  
  et_node_struct(et_node en);
  // Create a new et_node_struct which is an inactive copy of en.

  // Destructor
  ~et_node_struct() {};
  
  // --- Conversion Functions ---
  et_node parent() { return (et_node) hdt_rnbw_node_struct::parent(); }
  et_node left_child() { return (et_node) child[rnb_left]; }
  et_node right_child() { return (et_node) child[rnb_right]; }
  et_tree find_root() { return (et_tree) rnb_node_struct::find_root(); }
  et_node first() { return (et_node) rnb_node_struct::first(); }
  et_node last() { return (et_node) rnb_node_struct::last(); }
  et_node cyclic_succ() { return (et_node) rnb_node_struct::cyclic_succ(); }
  et_node cyclic_pred() { return (et_node) rnb_node_struct::cyclic_pred(); }
  friend inline et_tree et_join(et_tree t1, et_tree t2, et_node dummy)
  { return (et_tree) rnb_join(t1,t2,dummy); }
  //  friend inline et_node et_locate(et_tree et, int w, int& offset)
  //  { return (et_node) hdt_rnbw_locate(et,w,offset); }
  
  node get_corr_node() { return corr_node; }
  // This et_node is an occurrence of the returned graph node.
  
  int is_active() { return active; }
  // true <=> active occ.
  
  friend et_tree et_link(node u, node v, edge e, int i, hdt_base* dc);
  // Modify the et_trees of dc at level i corresponding to the insertion of
  // the edge (u,v) into F_i.
  // Prec.: u and v belong to dc, and they are not connected at the
  //        valid level i.
  
  friend void et_cut(edge e, int i, hdt_base* dc);
  // Update the et_trees at level i corresponding to the removal of
  // the tree edge e.
  // Prec.: e actually is a tree edge at level i.
  
  void print();
  // Print this node to cout for testing.

  friend void move_tree_edges( et_tree root, int i, hdt_base *dc );
  // in file dc_hdt.c, so it can access dc pointer

  friend void move_tree_edges_traversal(et_tree root, int i, hdt_base *dc);

  friend bool find_replacement_edge( et_tree root, int i, hdt_base *dc );
  // in file dc_hdt.c, so it can access dc pointer

  friend bool samp_for_replacement( et_tree root, int i, hdt_base *dc, 
				    int samp_threshold );
  // in file dc_hdt.c, so it can access dc pointer

protected:

  hdt_base*  dc;          // the dynamic connectivity data structure this node
                         // belongs to
  node      corr_node;   // corresponding node in G
  int       level;       // the level of this node
  int       active;      // true iff active occurrence
  edge      edge_occ[2]; // the at most two tree edges represented
                         // also by this node. ordered left and right

  void pass_activity(et_node to);
  // Make this node inactive and pass its activity to to.
  // Prec.: This node is active, to represents the same vertex and is on the
  //        same level.
  
  friend void change_root(et_tree& et, et_node en, int i, hdt_base* dc);
  // Change the root of the tree T represented by the et_tree et to the
  // vertex represented by the et_node en. The new tree is stored at et.
  // Prec.: The et_node en is in the et_tree et.
};


// ---------- node and edge data -------------- //
struct hdt_node_struct
{
    et_node *act_occ;		// array of active occurrences
  list<edge> *tree_edges;	// array of list of tree edges for each level
  list<edge> *non_tree_edges;  // array of list of non-tree edges for each 
				  //   level
	
	hdt_node_struct()
	{ act_occ = nil; tree_edges = nil; non_tree_edges = nil;}
};

struct hdt_edge_struct
{
  int        level;            // this edge belongs to G_level
  list_item  item[2];
  // could be a tree or non_tree item but only one of the  two.
  // So this is enough
  node end_point[2];
  // Store the endpoints corresponding to each of the list_items
  et_node  **tree_occ;         
  // tree edge: array for each level the
  // 4 pointers to occurrences repr. this edge
  // non-tree edge: nil
  
  hdt_edge_struct()
  { 
    item[0] = nil; item[1] = nil;
    end_point[0] = end_point[1] = nil;
    tree_occ = nil; 
  }
};

typedef hdt_node_struct* hdt_node_inf;
typedef hdt_edge_struct* hdt_edge_inf;

typedef GRAPH<hdt_node_inf, hdt_edge_inf> dc_graph;

class hdt_base{

public:

  hdt_base(dc_graph& G, int n_levels = -1, int samp_threshold_edges = -1, int base_size = -1);
	  //: initialized(false)
	  //{ init(G, n_levels); }
  // constructor, initializes the dynamic connectivity data structure
  // if n_levels > 0 then it specifies the number of levels (default O(log n))

	  //hdt_base()  { initialized = false;}
  // constructor taking no arguments, you have to use init before
  // doing anything with an instance created by this constructor

	  virtual ~hdt_base();// { clear(); }
  // destructor

	  //void init(dc_graph& H, int n_levels = -1);
  // initialize the data structure for the graph H
  // if n_levels > 0 then it specifies the number of levels (default O(log n))

	  //void ins(edge e);
  // insert the existing edge e into the data structure

  edge ins(node u, node v);
  // create an edge connecting u and v and return it

  // void remove(edge e);
  // remove edge e from the data structure, but not from the graph

  void del(edge e);
  // delete the edge e in the data structure and in the graph

  bool connected(node u, node v);
  // return true if u and v are connected in the current graph
  // and false otherwise

  bool connected() { return (components <= 1); }
  // returns true if the graph is connected

  int  number_of_components() { return components; }
  // returns the current number of components

  void print_statistics(ostream& out);
  // prints some statistics to the output stream out

protected:
  // global data
  dc_graph *Gp;
  int max_level;
  et_node et_dummy;

  bool initialized;
  int components;
  
  // node and edge related data
  // node_map<hdt_node_inf> node_info;
  //edge_map<hdt_edge_inf> edge_info;
	
  int n_ins;           // number of ins operations
  int n_del;           // number of del operations
  int n_query;         // number of user supplied connected queries
  int n_connected;     // number of conecteds
  int n_ins_tree;      // number of insert_trees
  int n_del_tree;      // number of delete_trees
  int n_ins_non_tree;  // number of insert_non_tree
  int n_del_non_tree;  // number of delete_non_tree
  int n_replace;       // number of replaces
  int n_replace_found; // number of successful replaces
  int n_tree_moved_up;// number of tree edges moved up in replace
  int n_nontree_moved_up;// number of nontree edges moved up in replace

  unsigned int *n_nontree_deletes_at_level;
  unsigned int *n_tree_deletes_at_level;

  //	int *n_added_edges;  // number of added edges
  
  // --- Internal Functions of the dynamic connectivity data structure --- //
  // Let G_i be the subgraph of G on level i, let F_i be the
  // forest in G_i, and let F be the spanning forest of G.
  // Let T be a spanning tree on level i.
  
  bool connected(node x, node y, int i);
  // Return true if x and y are connected on level i. Otherwise
  // return false.
  
  bool tree_edge(edge e)
    // Return true if e is an edge in F, false otherwise.
  { return ((*Gp)[e]->tree_occ != nil); }
  
  int level(edge e)
    // Return i such that e is in G_i.
  { return (*Gp)[e]->level; }
  
  void insert_tree(edge e, int i, bool create_tree_occs = false);
  // Insert e into F_i. If create_tree_occs is true, then the storage for
  // the tree_occ array for e is allocated.
  
  void delete_tree(edge e);
  // Remove the tree edge e from F.
  
  void replace(node u, node v, int i);
  // Replace the deleted tree edge (u,v) at level i.
  
  void insert_non_tree(edge e, int i);
  // Insert the non-tree edge e into G_i.
  
  void delete_non_tree(edge e);
  // Delete the non-tree edge e.

  friend void move_tree_edges( et_tree root, int i, hdt_base *dc );
  // used by replace:  move all level i tree edges incident to tree
  // to i+1, if any exist

  friend void move_tree_edges_traversal(et_tree root, int i, hdt_base *dc);
  // used by replace:  move all level i tree edges incident to tree
  // to i+1, if any exist

  friend bool find_replacement_edge( et_tree root, int i, hdt_base *dc );
  // used by replace:  find replacement nontree edge of level i
  // if it exists

  friend bool samp_for_replacement( et_tree root, int i, hdt_base *dc, 
				    int samp_threshold );
  // in file dc_hdt.c, so it can access dc pointer

  void clear();
  // deallocates all additional memory allocated for the data structure.

  friend class et_node_struct;
  friend void change_root(et_tree& et, et_node en, int i, hdt_base* dc);
  friend et_tree et_link(node u, node v, edge e, int i, hdt_base* dc);
  friend void et_cut(edge e, int i, hdt_base* dc);
};

#endif  /* _DC_HDT_H */

